from typing import List, Dict

from flotorch_core.chunking.chunking import Chunk
from .bedrock_embedding import BedRockEmbedding
from .embedding_registry import register


@register("amazon.titan-embed-image-v1")
@register("amazon.titan-embed-text-v1")
@register("amazon.titan-text-express-v1")
class TitanV1Embedding(BedRockEmbedding):
    """
    Initializes the TitanV1Embedding class.
    :param model_id: The model id of the TitanV1 model.
    :param region: The region of the TitanV1 model.
    :param dimensions: The dimensions of the embedding.
    :param normalize: Normalize the embedding.
    """

    def __init__(self, model_id: str, region: str, dimensions: int = 256, normalize: bool = True) -> None:
        super().__init__(model_id, region, dimensions, normalize)

    """
    Prepares the chunk for embedding.
    :param chunk: The chunk to be embedded.
    :return: The prepared chunk.
    """
    def _prepare_chunk(self, chunk: Chunk) -> Dict:
        return {"inputText": chunk.data, "embeddingConfig": {"outputEmbeddingLength": self.dimension}}

    """
    Extracts the embedding from the response.
    :param response: The response from the model.
    :return: The embedding.
    """
    def extract_embedding(self, response: Dict) -> List[float]:
        return response["embedding"]



import time
import unittest
from unittest.mock import MagicMock, patch
from botocore.exceptions import ClientError
from flotorch_core.utils.boto_retry_handler import BotoRetryHandler, RetryParams
from flotorch_core.utils.bedrock_retry_handler import BedRockRetryHander

class TestBotoRetryHandler(unittest.TestCase):
    
    def setUp(self):
        self.retry_handler = BedRockRetryHander()
    
    @patch("time.sleep", return_value=None)  # Patch sleep to speed up tests
    def test_successful_call(self, mock_sleep):
        mock_func = MagicMock(return_value="Success")
        wrapped_func = self.retry_handler(mock_func)
        result = wrapped_func()
        self.assertEqual(result, "Success")
        mock_func.assert_called_once()
    
    @patch("time.sleep", return_value=None)
    def test_retryable_error(self, mock_sleep):
        mock_func = MagicMock()
        error_response = {"Error": {"Code": "ThrottlingException"}}
        mock_func.side_effect = [ClientError(error_response, "TestOp")] * 3 + ["Success"]
        
        wrapped_func = self.retry_handler(mock_func)
        result = wrapped_func()
        
        self.assertEqual(result, "Success")
        self.assertEqual(mock_func.call_count, 4)  # 3 failures + 1 success
    
    @patch("time.sleep", return_value=None)
    def test_max_retries_exceeded(self, mock_sleep):
        mock_func = MagicMock()
        error_response = {"Error": {"Code": "ThrottlingException"}}
        mock_func.side_effect = ClientError(error_response, "TestOp")
        
        wrapped_func = self.retry_handler(mock_func)
        
        with self.assertRaises(ClientError):
            wrapped_func()
        
        self.assertEqual(mock_func.call_count, self.retry_handler.retry_params.max_retries)
    
    @patch("time.sleep", return_value=None)
    def test_non_retryable_error(self, mock_sleep):
        mock_func = MagicMock()
        error_response = {"Error": {"Code": "AccessDeniedException"}}  # Not in retryable_errors
        mock_func.side_effect = ClientError(error_response, "TestOp")
        
        wrapped_func = self.retry_handler(mock_func)
        
        with self.assertRaises(ClientError):
            wrapped_func()
        
        self.assertEqual(mock_func.call_count, 1)  # Should not retry for non-retryable errors
    
    @patch("time.sleep", return_value=None)
    def test_unexpected_exception(self, mock_sleep):
        mock_func = MagicMock(side_effect=ValueError("Unexpected error"))
        wrapped_func = self.retry_handler(mock_func)
        
        with self.assertRaises(ValueError):
            wrapped_func()
        
        self.assertEqual(mock_func.call_count, 1)  # Should not retry for unexpected errors

if __name__ == "__main__":
    unittest.main()



import random
import time
from typing import Any, Dict, List, Tuple
import boto3
from flotorch_core.inferencer.inferencer import BaseInferencer
from flotorch_core.logger.global_logger import get_logger
from flotorch_core.utils.sagemaker_utils import SageMakerUtils, INFERENCER_MODELS
from sagemaker.session import Session
from sagemaker.serializers import JSONSerializer
from sagemaker.deserializers import JSONDeserializer
from sagemaker.predictor import Predictor

logger = get_logger()

class SageMakerInferencer(BaseInferencer):
    def __init__(self, model_id: str, region: str, role_arn: str, n_shot_prompts: int = 0, temperature: float = 0.7, n_shot_prompt_guide_obj: Dict[str, List[Dict[str, str]]] = None):
        """
        Initialize the BedrockInferencer with Bedrock-specific parameters.

        Args:
            model_id (str): Identifier for the Bedrock model.
            region (str): AWS region where the Bedrock service is deployed.
            n_shot_prompts (int): Number of examples to include in few-shot learning.
            temperature (float): Sampling temperature for response generation.
            n_shot_prompt_guide_obj (Dict[str, List[Dict[str, str]]]): Guide object for few-shot examples.
        """
        super().__init__(model_id, region, n_shot_prompts, temperature, n_shot_prompt_guide_obj)
        self.role = role_arn
        self.client = boto3.client("sagemaker-runtime", region_name=region)
        self.sagemaker_client = boto3.client('sagemaker', region_name=region)

        logger.info(f"Initializing SageMaker Generator for model: {model_id}")

        self.session = Session(boto_session=boto3.Session(region_name=region))

        self.inferencing_model_id = model_id
        self.inferencing_model_endpoint_name = f"{SageMakerUtils.sanitize_name(model_id)[:42]}-inferencing-endpoint"

        model_config = INFERENCER_MODELS.get(model_id)
        if not SageMakerUtils.check_endpoint_exists(self.sagemaker_client, self.inferencing_model_endpoint_name):
            if INFERENCER_MODELS[model_id]['model_source'] == 'jumpstart':
                SageMakerUtils.create_jumpstart_endpoint(self.sagemaker_client, model_config.get("instance_type"), self.region_name, self.role, model_id, self.inferencing_model_endpoint_name)
            elif INFERENCER_MODELS[model_id]['model_source'] == 'huggingface':
                SageMakerUtils.create_huggingface_endpoint(self.sagemaker_client, model_config.get("instance_type"), model_id, self.inferencing_model_endpoint_name, self.role, self.region_name)

        SageMakerUtils.wait_for_endpoint_creation(self.sagemaker_client, self.inferencing_model_endpoint_name)


        self.predictor = Predictor(
            endpoint_name=self.inferencing_model_endpoint_name,
            sagemaker_session=self.session
        )

        self.predictor.serializer = JSONSerializer()
        self.predictor.deserializer = JSONDeserializer()

        self.inferencing_predictor = self.predictor

    def generate_text(self, user_query: str, context: List[Dict]) -> Tuple[Dict[Any, Any], str]:
        if not self.inferencing_predictor:
            raise ValueError("Generation predictor not initialized")
        
        system_prompt, prompt = self.generate_prompt(user_query, context)

        payload = self.construct_payload(system_prompt, prompt)

        try:
            start_time = time.time()
            response = self.inferencing_predictor.predict(payload)
            latency = int((time.time() - start_time) * 1000)

            generated_text = self._extract_response(response)
            
            if "The final answer is:" in generated_text:
                answer = generated_text.split("The final answer is:")[1].strip()
            elif "Assistant:" in generated_text:
                answer = generated_text.split("Assistant:")[1].strip()
            else:
                answer = generated_text.strip()

            cleaned_response = self._clean_response(answer)

            if not cleaned_response or cleaned_response.isspace() or 'DRAFT' in cleaned_response:
                return None, "Unable to generate a proper response. Please try again."
            
            input_tokens = len(prompt) // 4
            output_tokens = len(generated_text) // 4
            total_tokens = input_tokens + output_tokens
            
            answer_metadata = {
                'inputTokens': input_tokens,
                'outputTokens': output_tokens,
                'totalTokens': total_tokens,
                'latencyMs': latency
            }
            
            return answer_metadata, cleaned_response
            

        except Exception as e:
            logger.error(f"Error generating response: {str(e)}")
            return f"Error generating response: {str(e)}"

    def _clean_response(self, text: str) -> str:
        """
        Cleans and formats the response text by removing common artifacts, ensuring proper sentence structure,
        and eliminating excessive whitespace or newlines.

        Args:
            text (str): The raw response text that needs to be cleaned.

        Returns:
            str: The cleaned and formatted response text.
        """
        artifacts = ['DRAFT', '[INST]', '[/INST]', 'Human:', 'Assistant:']
        
        cleaned_text = text.strip()
        for artifact in artifacts:
            cleaned_text = cleaned_text.replace(artifact, '').strip()

        sentence_endings = ['.', '!', '?']
        
        if not any(cleaned_text.rstrip().endswith(end) for end in sentence_endings):
            last_period = max(
                cleaned_text.rfind('.'),
                cleaned_text.rfind('!'),
                cleaned_text.rfind('?')
            )
            
            if last_period != -1:
                cleaned_text = cleaned_text[:last_period + 1]

        cleaned_text = ' '.join(cleaned_text.split())
        
        think_end_index = cleaned_text.find('</think>')
        if think_end_index != -1:
            cleaned_text = cleaned_text[think_end_index + len('</think>'):]

        return cleaned_text.strip()            

    def generate_prompt(self, user_query: str, context: List[Dict]) -> Tuple[str, List[Dict[str, Any]]]:
        if self.n_shot_prompts < 0:
            raise ValueError("n_shot_prompt must be non-negative")
        
        default_prompt = "You are a helpful assistant. Use the provided context to answer questions accurately. If you cannot find the answer in the context, say so"
        # Get system prompt
        system_prompt = default_prompt if not self.n_shot_prompt_guide_obj or not self.n_shot_prompt_guide_obj.get("system_prompt") else self.n_shot_prompt_guide_obj.get("system_prompt")
        
        context_text = self.format_context(user_query, context)

        base_prompt = self.n_shot_prompt_guide_obj.get("user_prompt", "") if self.n_shot_prompt_guide_obj else ""

        if self.n_shot_prompts == 0:
            logger.info("into zero shot prompt")

            if self.inferencing_model_id == "huggingface-llm-falcon-7b-instruct-bf16":
                prompt = f"""Below are search results and a query. Create a concise summary.
                    Query: {user_query}
                    Search Results: {context_text}
                    Summary:"""
                return None, prompt
            
            prompt = f"Human: {system_prompt}\n\n{context_text}\n\n{base_prompt}\n\nAssistant: The final answer is:"
            return None, prompt.strip()
        
        examples = self.n_shot_prompt_guide_obj['examples']
        selected_examples = (random.sample(examples, self.n_shot_prompts) 
                        if len(examples) > self.n_shot_prompts 
                        else examples)
        
        example_text = ""
        for example in selected_examples:
            example_text += f"- {example['example']}\n"

        logger.info(f"into {self.n_shot_prompts} shot prompt  with examples {len(selected_examples)}")

        if self.inferencing_model_id == "huggingface-llm-falcon-7b-instruct-bf16":
            prompt = f"""Below are search results and a query. Create a concise summary.
                Query: {user_query}
                Few examples:\n 
                {example_text}\n 
                Search Results: {context_text}
                Summary:"""
                
            return None, prompt
        
        prompt = f"Human: {system_prompt}\n\nFew examples:\n{example_text}\n{context_text}\n\n{base_prompt}\n\nAssistant: The final answer is:"
        return prompt.strip()
    
    def format_context(self, user_query: str, context: List[Dict[str, str]]) -> str:
        """Format context documents into a single string."""
        formatted_context = f"Search Query: {user_query}\n"
        
        try:
            if context:
                formatted_context += "\nRelevant Passages:\n"
                for i, item in enumerate(context, 1):
                    content = None
                    if 'text' in item:
                        content = item['text']
                    elif '_source' in item and 'text' in item['_source']:
                        content = item['_source']['text']
                    
                    if not content:
                        continue

                    score = item.get('_score', 'N/A')
                    formatted_context += f"\nPassage {i} (Score: {score}):\n{content}\n"
            return formatted_context
        except Exception as e:
            logger.error(f"Error formatting context: {str(e)}")
            formatted_context += "Error processing context"
            return formatted_context
        
    def construct_payload(self, system_prompt: str, prompt: str) -> dict:
        """
        Constructs the payload dictionary for model inference with the given prompts and default parameters.
        
        Args:
            system_prompt (str): The system-level prompt that guides the model's behavior
            prompt (str): The actual prompt/query to be sent to the model

        """
        # Define default parameters for controlling the model's text generation
        default_params = {
            "max_new_tokens": 256,
            "temperature": self.temperature,
            "top_p": 0.9,
            "do_sample": True
            }
        # Construct the complete payload with prompt and generation parameters
        payload = {
            "inputs": prompt,
            "parameters": default_params
            }
        
        return payload
    
    def _extract_response(self, response: dict) -> str:
        """
        Parses the response from the model and extracts the generated text.

        Args:
            response (dict): The raw response from the model
        """
        # Handle different response formats (Falcon vs Llama)
        if isinstance(response, list):
            # Falcon-style response: Retrieve generated text from the list
            return response[0].get('generated_text', '') if response else ''
        elif isinstance(response, dict):
            return response.get('generated_text', '')
        else:
            raise ValueError(f"Unexpected response format: {type(response)}")


from .logger_provider import LoggerProvider
import logging

class ConsoleLoggerProvider(LoggerProvider):
    """
    Logger provider that logs messages to the console.
    """

    def __init__(self, name: str = "default"):
        self.logger = logging.getLogger(name)
        self.logger.setLevel(logging.INFO)

        if not self.logger.handlers:  
            handler = logging.StreamHandler()
            formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
            handler.setFormatter(formatter)
            self.logger.addHandler(handler)

    def log(self, level: str, message: str) -> None:
        log_method = getattr(self.logger, level.lower(), self.logger.info)
        log_method(message)

    def get_logger(self) -> logging.Logger:
        return self.logger




import unittest
from unittest.mock import MagicMock, patch
from botocore.exceptions import ClientError
from pydantic import ValidationError

from flotorch_core.utils.boto_retry_handler import BotoRetryHandler, RetryParams

class TestRetryParams(unittest.TestCase):
    """Tests for the RetryParams model"""
    
    def test_valid_retry_params(self):
        params = RetryParams(max_retries=3, retry_delay=1, backoff_factor=2)
        self.assertEqual(params.max_retries, 3)
        self.assertEqual(params.retry_delay, 1)
        self.assertEqual(params.backoff_factor, 2)
    
    def test_invalid_retry_params(self):
        with self.assertRaises(ValidationError):
            RetryParams(max_retries="invalid", retry_delay=1, backoff_factor=2)
        
        with self.assertRaises(ValidationError):
            RetryParams(max_retries=3, retry_delay="invalid", backoff_factor=2)
        
        with self.assertRaises(ValidationError):
            RetryParams(max_retries=3, retry_delay=1, backoff_factor="invalid")


class ConcreteBotoRetryHandler(BotoRetryHandler):
    """A concrete implementation of BotoRetryHandler for testing"""
    
    @property
    def retry_params(self) -> RetryParams:
        return RetryParams(max_retries=3, retry_delay=1, backoff_factor=2)
    
    @property
    def retryable_errors(self) -> set[str]:
        return {"ThrottlingException", "RequestLimitExceeded"}


class TestBotoRetryHandler(unittest.TestCase):
    """Tests for the abstract BotoRetryHandler class"""
    
    def setUp(self):
        self.retry_handler = ConcreteBotoRetryHandler()
    
    def test_abstract_class_instantiation(self):
        """Test that direct instantiation of the abstract class raises TypeError"""
        with self.assertRaises(TypeError):
            BotoRetryHandler()
    
    @patch("time.sleep", return_value=None)  # Patch sleep to speed up tests
    def test_successful_call(self, mock_sleep):
        mock_func = MagicMock(return_value="Success")
        wrapped_func = self.retry_handler(mock_func)
        result = wrapped_func()
        self.assertEqual(result, "Success")
        mock_func.assert_called_once()
        mock_sleep.assert_not_called()
    
    @patch("time.sleep", return_value=None)
    def test_retryable_error(self, mock_sleep):
        mock_func = MagicMock()
        error_response = {"Error": {"Code": "ThrottlingException"}}
        mock_func.side_effect = [
            ClientError(error_response, "TestOp"),
            "Success"
        ]
        
        wrapped_func = self.retry_handler(mock_func)
        result = wrapped_func()
        
        self.assertEqual(result, "Success")
        self.assertEqual(mock_func.call_count, 2)  # 1 failure + 1 success
        mock_sleep.assert_called_once_with(1)  # First retry delay
    
    @patch("time.sleep", return_value=None)
    def test_multiple_retries(self, mock_sleep):
        mock_func = MagicMock()
        error_response = {"Error": {"Code": "ThrottlingException"}}
        mock_func.side_effect = [
            ClientError(error_response, "TestOp"),
            ClientError(error_response, "TestOp"),
            "Success"
        ]
        
        wrapped_func = self.retry_handler(mock_func)
        result = wrapped_func()
        
        self.assertEqual(result, "Success")
        self.assertEqual(mock_func.call_count, 3)  # 2 failures + 1 success
        # Check that sleep was called with the correct backoff times
        mock_sleep.assert_any_call(1)  # First retry
        mock_sleep.assert_any_call(2)  # Second retry with backoff
    
    @patch("time.sleep", return_value=None)
    def test_max_retries_exceeded(self, mock_sleep):
        mock_func = MagicMock()
        error_response = {"Error": {"Code": "ThrottlingException"}}
        
        # Create a ClientError object that we'll use consistently
        client_error = ClientError(error_response, "TestOp")
        
        # Set up the mock to always raise the same error
        mock_func.side_effect = client_error
        
        wrapped_func = self.retry_handler(mock_func)
        
        # We expect the function to raise the same ClientError after max retries
        with self.assertRaises(ClientError) as context:
            wrapped_func()
            
        # Verify it's the same error object
        self.assertIs(context.exception, client_error)
        
        # Should try exactly max_retries times
        self.assertEqual(mock_func.call_count, self.retry_handler.retry_params.max_retries)
        
        # Check sleep was called with correct backoff times
        self.assertEqual(mock_sleep.call_count, self.retry_handler.retry_params.max_retries - 1)
        mock_sleep.assert_any_call(1)  # First retry
        mock_sleep.assert_any_call(2)  # Second retry with backoff
        # No need to check for the third call since we're only doing max_retries=3
    
    @patch("time.sleep", return_value=None)
    def test_different_retryable_error(self, mock_sleep):
        mock_func = MagicMock()
        error_response = {"Error": {"Code": "RequestLimitExceeded"}}  # Different retryable error
        mock_func.side_effect = [
            ClientError(error_response, "TestOp"),
            "Success"
        ]
        
        wrapped_func = self.retry_handler(mock_func)
        result = wrapped_func()
        
        self.assertEqual(result, "Success")
        self.assertEqual(mock_func.call_count, 2)
    
    @patch("time.sleep", return_value=None)
    def test_non_retryable_error(self, mock_sleep):
        mock_func = MagicMock()
        error_response = {"Error": {"Code": "AccessDeniedException"}}  # Not in retryable_errors
        mock_func.side_effect = ClientError(error_response, "TestOp")
        
        wrapped_func = self.retry_handler(mock_func)
        
        with self.assertRaises(ClientError):
            wrapped_func()
        
        self.assertEqual(mock_func.call_count, 1)  # Should not retry for non-retryable errors
        mock_sleep.assert_not_called()
    
    @patch("time.sleep", return_value=None)
    def test_unexpected_exception(self, mock_sleep):
        mock_func = MagicMock(side_effect=ValueError("Unexpected error"))
        wrapped_func = self.retry_handler(mock_func)
        
        with self.assertRaises(ValueError):
            wrapped_func()
        
        self.assertEqual(mock_func.call_count, 1)  # Should not retry for unexpected errors
        mock_sleep.assert_not_called()
    
    @patch("flotorch_core.utils.boto_retry_handler.logger")  # Patch logger directly
    @patch("time.sleep", return_value=None)
    def test_logging(self, mock_sleep, mock_logger):
        # Mock function that raises a retryable error first
        mock_func = MagicMock()
        error_response = {"Error": {"Code": "ThrottlingException"}}
        mock_func.side_effect = [
            ClientError(error_response, "TestOp"),
            "Success"
        ]

        # Call wrapped function
        wrapped_func = self.retry_handler(mock_func)
        result = wrapped_func()

        # Ensure function succeeded
        self.assertEqual(result, "Success")

        # Verify that logger.error was called
        mock_logger.error.assert_called()  
        mock_logger.info.assert_any_call("Retrying in 1 seconds...")

if __name__ == "__main__":
    unittest.main()



from flotorch_core.chunking.chunking import Chunk
from flotorch_core.storage.db.vector.vector_storage import VectorStorage, VectorStorageSearchResponse
from flotorch_core.guardrails.guardrails import BaseGuardRail

class GuardRailsVectorStorage(VectorStorage):

    def __init__(self, vectorStorage: VectorStorage, base_guardrail: BaseGuardRail, 
                 apply_prompt=False, apply_context=False):
        self.vectorStorage = vectorStorage
        self.base_guardrail = base_guardrail
        self.apply_prompt = apply_prompt
        self.apply_context = apply_context
    
    def search(self, chunk: Chunk, knn: int, hierarchical=False):
        if self.apply_prompt:
            guardrail_response = self.base_guardrail.apply_guardrail(chunk.data, 'INPUT')
            if guardrail_response['action'] == 'GUARDRAIL_INTERVENED':
                return VectorStorageSearchResponse(
                    status=False,
                    metadata={
                        'guardrail_output': guardrail_response['outputs'][0]['text'],
                        'guardrail_input_assessment': guardrail_response.get('assessments', []),
                        'block_level': 'INPUT',
                        'guardrail_blocked': True
                    }
                )
            
            results = self.vectorStorage.search(chunk, knn, hierarchical)

        if self.apply_context:
            result_text = ' '.join(record.text for record in results.result)
            guardrail_response = self.base_guardrail.apply_guardrail(result_text, 'INPUT')
            if guardrail_response['action'] == 'GUARDRAIL_INTERVENED':
                return VectorStorageSearchResponse(
                    status=False,
                    result=results.result,
                    metadata={
                        'guardrail_output': guardrail_response['outputs'][0]['text'],
                        'guardrail_context_assessment': guardrail_response.get('assessments', []),
                        'block_level': 'CONTEXT',
                        'guardrail_blocked': True,
                        'embedding_metadata': results.metadata['embedding_metadata'] if 'embedding_metadata' in results.metadata else {}
                    }
                )
            
        return results
    
    def embed_query(self, embedding, knn, hierarical=False):
        self.vectorStorage.embed_query(embedding, knn, hierarical)

    def write(self, body):
        self.vectorStorage.write(body)

    def read(self, body):
        self.vectorStorage.read(body)


import logging
import pytest
from flotorch_core.logger.global_logger import get_logger
from flotorch_core.logger.logger import Logger
from flotorch_core.logger.console_logger_provider import ConsoleLoggerProvider

def test_get_logger_returns_logger_instance():
    """Test that get_logger() returns a Logger instance."""
    logger = get_logger()
    assert isinstance(logger, Logger)

def test_get_logger_uses_console_logger_provider():
    """Test that get_logger() initializes with ConsoleLoggerProvider."""
    logger = get_logger()
    assert isinstance(logger.provider, ConsoleLoggerProvider)

def test_get_logger_singleton():
    """Test that get_logger() returns the same instance (singleton)."""
    logger1 = get_logger()
    logger2 = get_logger()
    assert logger1 is logger2  # Ensure both are the same instance

def test_get_logger_logs_messages(caplog):
    """Test that logging through get_logger() works."""
    logger = get_logger()
    
    with caplog.at_level(logging.INFO):
        logger.log("info", "Test singleton logger message")
    
    assert "Test singleton logger message" in caplog.text



import random
from openai import OpenAI
from typing import List, Dict, Tuple

from flotorch_core.inferencer.inferencer import BaseInferencer
import time


class GatewayInferencer(BaseInferencer):
    def __init__(self, model_id: str, api_key: str, base_url: str = None, n_shot_prompts: int = 0, n_shot_prompt_guide_obj: Dict[str, List[Dict[str, str]]] = None):
        super().__init__(model_id, None, n_shot_prompts, None, n_shot_prompt_guide_obj)
        self.api_key = api_key
        self.base_url = base_url
        self.client = OpenAI(api_key=self.api_key, base_url=self.base_url)

    def generate_prompt(self, user_query: str, context: List[Dict]) -> List[Dict[str, str]]:
        messages = []
        
        # System prompt
        default_prompt = "You are a helpful assistant. Use the provided context to answer questions accurately. If you cannot find the answer in the context, say so"
        system_prompt = (
            self.n_shot_prompt_guide_obj.get("system_prompt", default_prompt)
            if self.n_shot_prompt_guide_obj
            else default_prompt
        )
        messages.append({"role": "system", "content": system_prompt})

        # Nshot examples
        if self.n_shot_prompt_guide_obj:
            examples = self.n_shot_prompt_guide_obj.get("examples", [])
            selected_examples = (
                random.sample(examples, self.n_shot_prompts)
                if len(examples) > self.n_shot_prompts
                else examples
            )
            for example in selected_examples:
                if "example" in example:
                    messages.append({"role": "assistant", "content": example["example"]})
                elif "question" in example and "answer" in example:
                    messages.append({"role": "user", "content": example["question"]})
                    messages.append({"role": "assistant", "content": example["answer"]})
             
        # Context
        if context:
            context_text = self.format_context(context)
            if context_text:
                messages.append({"role": "user", "content": context_text})

        # User query and base prompt
        base_prompt = self.n_shot_prompt_guide_obj.get("user_prompt", "") if self.n_shot_prompt_guide_obj else ""
        # Combine base prompt with user query if base prompt is provided else use user query
        query = base_prompt + "\n" + user_query if base_prompt else user_query
        messages.append({"role": "user", "content": query})
        
        return messages

    def generate_text(self, user_query: str, context: List[Dict]) -> Tuple[Dict, str]:
        messages  = self.generate_prompt(user_query, context)
        
        start_time = time.time()
        response = self.client.chat.completions.create(
            model=self.model_id,
            messages=messages
        )


        end_time = time.time()

        metadata = self._extract_metadata(response)
        metadata["latencyMs"] = str(int((end_time - start_time) * 1000))
        
        return metadata, response.choices[0].message.content


    def format_context(self, context: List[Dict[str, str]]) -> str:
        """
        Format context into a string to be included in the prompt.
        """
        return "\n".join([f"Context {i+1}:\n{item['text']}" for i, item in enumerate(context)])
    
    def _extract_metadata(self, response):
        return {
            "inputTokens": str(response.usage.prompt_tokens),
            "outputTokens": str(response.usage.completion_tokens),
            "totalTokens": str(response.usage.total_tokens),
            "latencyMs": "0"
        }


import unittest
from unittest.mock import patch, mock_open

from storage.local_storage import LocalStorageProvider
from storage.s3_storage import S3StorageProvider
from testcontainers.minio import MinioContainer
import boto3
from moto import mock_aws


class TestLocalStorageProvider(unittest.TestCase):

    @patch('builtins.open', new_callable=mock_open)
    @patch('os.path.isdir', return_value=False)
    def test_write(self, mock_isdir, mock_open):
        provider = LocalStorageProvider()
        provider.write('test_path', 'test_data')
        mock_open.assert_called_once_with('test_path', 'wb')
        mock_open().write.assert_called_once_with('test_data')

    @patch('builtins.open', new_callable=mock_open, read_data='test_data')
    @patch('os.path.isdir', return_value=False)
    def test_read(self, mock_isdir, mock_open):
        provider = LocalStorageProvider()
        result = list(provider.read('test_path'))
        mock_open.assert_called_once_with('test_path', 'rb')
        self.assertEqual(result, ['test_data'])


class TestS3StorageProvider(unittest.TestCase):

    @patch('boto3.client')
    def setUp(self, mock_boto_client):
        self.mock_s3_client = mock_boto_client.return_value
        self.provider = S3StorageProvider('test_bucket', s3_client=self.mock_s3_client)

    def test_write(self):
        self.provider.write('/tmp/test_path.data', 'test_data')
        self.mock_s3_client.put_object.assert_called_once_with(Bucket='test_bucket', Key='/tmp/test_path.data',
                                                               Body='test_data')

# class TestS3StorageContainer(unittest.TestCase):
    # Use testcontainer and test the S3StorageProvider
    class TestS3StorageContainer(unittest.TestCase):

        @mock_aws
        def setUp(self):
            self.minio_container = MinioContainer()
            self.minio_container.start()
            
            self.s3_client = boto3.client(
                's3',
                endpoint_url=self.minio_container.get_url(),
                aws_access_key_id=self.minio_container.access_key,
                aws_secret_access_key=self.minio_container.secret_key,
                region_name='us-east-1'
            )
            
            self.s3_client.create_bucket(Bucket='test_bucket')
            self.provider = S3StorageProvider('test_bucket', s3_client=self.s3_client)

        @mock_aws
        def tearDown(self):
            self.minio_container.stop()

        def test_write(self):
            self.provider.write('/tmp/test_path.data', 'this is a test data')


# class TestPDFReader(unittest.TestCase):
#
#     @patch('storage.storage.PdfReader')
#     def test_read_pdf(self, mock_pdf_reader):
#         mock_pdf_reader.return_value.pages = [MagicMock(extract_text=lambda: 'page_text')]
#         mock_storage_provider = MagicMock()
#         mock_storage_provider.read.return_value = [b'pdf_data']
#         reader = PDFReader(mock_storage_provider)
#         result = reader.read_pdf('test_path')
#         self.assertEqual(result, ['page_text'])

if __name__ == '__main__':
    unittest.main()



from typing import Dict, Optional, Any
import yaml


class GuardrailCreateConfig:
    def __init__(
            self,
            name: str,
            description: str,
            content_policy: Optional[Dict[str, Any]] = None,
            topic_policy: Optional[Dict[str, Any]] = None,
            word_policy: Optional[Dict[str, Any]] = None,
            sensitive_info_policy: Optional[Dict[str, Any]] = None,
            contextual_grounding_policy: Optional[Dict[str, Any]] = None,
            input_filter: Optional[Dict[str, Any]] = None,
            output_filter: Optional[Dict[str, Any]] = None
    ):
        self.name = name
        self.description = description
        self.content_policy = content_policy
        self.topic_policy = topic_policy
        self.word_policy = word_policy
        self.sensitive_info_policy = sensitive_info_policy
        self.contextual_grounding_policy = contextual_grounding_policy
        self.input_filter = input_filter
        self.output_filter = output_filter

    def to_dict(self) -> Dict[str, Any]:
        return {
            "name": self.name,
            "description": self.description,
            "content_policy": self.content_policy,
            "topic_policy": self.topic_policy,
            "word_policy": self.word_policy,
            "sensitive_info_policy": self.sensitive_info_policy,
            "contextual_grounding_policy": self.contextual_grounding_policy,
            "input_filter": self.input_filter,
            "output_filter": self.output_filter
        }

    @staticmethod
    def from_yaml(yaml_file: str) -> 'GuardrailCreateConfig':
        with open(yaml_file, 'r') as file:
            config_data = yaml.safe_load(file)
        return GuardrailCreateConfig(
            name=config_data.get('name'),
            description=config_data.get('description'),
            content_policy=config_data.get('content_policy'),
            topic_policy=config_data.get('topic_policy'),
            word_policy=config_data.get('word_policy'),
            sensitive_info_policy=config_data.get('sensitive_info_policy'),
            contextual_grounding_policy=config_data.get('contextual_grounding_policy'),
            input_filter=config_data.get('input_filter'),
            output_filter=config_data.get('output_filter')
        )



from typing import List, Dict, Any, Optional
from decimal import Decimal
from datetime import datetime

class DBUtils:
    """
    Utility class for database operations
    """
    @staticmethod
    def prepare_item_for_write(item: Dict[str, Any]) -> Optional[Dict[str, Any]]:
        """
        Prepares a dictionary item for writing to PostgreSQL.
        Manually converts dictionary values to JSON strings.

        Args:
            item: The dictionary to prepare.

        Returns:
            A new dictionary with dict values converted to JSON strings,
            or None if a serialization error occurs.
        """
        item_prepared = {}
        try:
            for key, value in item.items():
                if isinstance(value, dict):
                    # Convert dictionaries to JSON strings
                    item_prepared[key] = json.dumps(value)
                else:
                    # Keep other types as they are
                    item_prepared[key] = value
            return item_prepared
        except TypeError as e:
            logger.error(f"Error during JSON serialization: {e}")
            return None
    
    @staticmethod
    def serialize_data(obj: Any) -> Any:
        """
        Recursively convert float to Decimal and serialize datetime to string.
        """
        if isinstance(obj, list):
            return [DBUtils.serialize_data(i) for i in obj]
        elif isinstance(obj, dict):
            return {k: DBUtils.serialize_data(v) for k, v in obj.items()}
        elif isinstance(obj, float):
            return Decimal(str(obj))  # Convert float to Decimal
        elif isinstance(obj, datetime):
            return obj.isoformat()  # Convert datetime to ISO 8601 string
        return obj


from typing import List, Dict
from flotorch_core.chunking.chunking import Chunk
from flotorch_core.config.config import Config
from flotorch_core.config.env_config_provider import EnvConfigProvider
from flotorch_core.embedding.sagemaker_embedding import SageMakerEmbedder
from flotorch_core.embedding.embedding_registry import register

env_config_provider = EnvConfigProvider()
config = Config(env_config_provider)



@register("huggingface-sentencesimilarity-bge-large-en-v1-5")
class BGELargeEmbedding(SageMakerEmbedder):
    """
    BGE Large Hugging Face model for sentence similarity.
    """
    def __init__(self, model_id: str, region: str, dimensions: int = 256, normalize: bool = True) -> None:
        super().__init__(model_id, region, config.get_sagemaker_arn_role(), dimensions, normalize)

    def _prepare_chunk(self, chunk: Chunk) -> Dict:
        return {"text_inputs": [chunk.data], "mode": "embedding"}


@register("huggingface-sentencesimilarity-bge-m3")
class BGEM3Embedding(SageMakerEmbedder):
    """
    BGE M3 Hugging Face model for sentence similarity.
    """
    def __init__(self, model_id: str, region: str, dimensions: int = 256, normalize: bool = True) -> None:
        super().__init__(model_id, region, config.get_sagemaker_arn_role(), dimensions, normalize)

    def _prepare_chunk(self, chunk: Chunk) -> Dict:
        return {"text_inputs": [chunk.data], "mode": "embedding"}


@register("huggingface-textembedding-gte-qwen2-7b-instruct")
class GTEQwen2Embedding(SageMakerEmbedder):
    """
    GTE Qwen2-7B Instruct Hugging Face model for text embedding.
    """
    def __init__(self, model_id: str, region: str, dimensions: int = 256, normalize: bool = True) -> None:
        super().__init__(model_id, region, config.get_sagemaker_arn_role(), dimensions, normalize)

    def _prepare_chunk(self, chunk: Chunk) -> Dict:
        return {"inputs": [chunk.data]}



from dataclasses import dataclass, field
import json
from flotorch_core.storage.db.db_storage import DBStorage
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional
from flotorch_core.embedding.embedding import BaseEmbedding

@dataclass
class VectorStorageSearchItem:
    text: str
    execution_id: Optional[str] = None
    chunk_id: Optional[str] = None
    parent_id: Optional[str] = None
    vectors: List[float] = field(default_factory=list)
    metadata: Dict[str, Any] = field(default_factory=dict)

    def to_json(self):
        return {
            "text": self.text,
            "execution_id": self.execution_id,
            "chunk_id": self.chunk_id,
            "parent_id": self.parent_id,
            "vectors": self.vectors,
            "metadata": self.metadata
        }
@dataclass
class VectorStorageSearchResponse:
    status: bool
    result: List[VectorStorageSearchItem] = field(default_factory=list)
    metadata: Dict[str, Any] = field(default_factory=dict)

    def to_json(self):
        return {
            "status": self.status,
            "result": [item.to_json() for item in self.result],
            "metadata": self.metadata
        }

    
class VectorStorage(DBStorage, ABC):
    def __init__(self, embedder: Optional[BaseEmbedding] = None):
        self.embedder = embedder
        
    
    @abstractmethod
    def search(self, query: str, knn, hierarchical=False):
        pass

    @abstractmethod
    def embed_query(self, embedding, knn, hierarical=False):
        pass
    



from typing import List, Dict

from flotorch_core.chunking.chunking import Chunk
from .embedding_registry import register
from .titanv1_embedding import TitanV1Embedding

"""
This class is responsible for embedding the text using the TitanV2 model.
"""
@register("amazon.titan-embed-text-v2:0")
class TitanV2Embedding(TitanV1Embedding):
    """
    Initializes the TitanV2Embedding class.
    :param model_id: The model id of the TitanV2 model.
    :param region: The region of the TitanV2 model.
    :param dimensions: The dimensions of the embedding.
    :param normalize: Normalize the embedding.
    """

    def __init__(self, model_id: str, region: str = "us-east-1", dimensions: int = 256, normalize: bool = True) -> None:
        super().__init__(model_id, region, dimensions, normalize)

    """
    Prepares the chunk for embedding.
    :param chunk: The chunk to be embedded.
    :return: The prepared chunk.
    """

    def _prepare_chunk(self, chunk: Chunk) -> Dict:
        return {"inputText": chunk.data, "dimensions": self.dimension, "normalize": self.normalize}



import pytest
from unittest.mock import Mock
from flotorch_core.logger.logger import Logger
from flotorch_core.logger.logger_provider import LoggerProvider

def test_logger_singleton():
    """Ensures Logger is a singleton."""
    provider_mock = Mock(spec=LoggerProvider)
    logger1 = Logger(provider_mock)
    logger2 = Logger()
    
    assert logger1 is logger2  # Both instances should be the same

def test_logger_requires_provider_on_first_initialization():
    """Ensures Logger requires a provider during first initialization."""
    Logger._instance = None  # Reset singleton for testing
    with pytest.raises(ValueError, match="LoggerProvider must be provided for the first initialization."):
        Logger()

def test_logger_delegates_logging():
    """Ensures Logger delegates log calls to the provider."""
    provider_mock = Mock(spec=LoggerProvider)
    Logger._instance = None  # Reset singleton
    logger = Logger(provider_mock)

    logger.log("INFO", "Test message")
    provider_mock.log.assert_called_once_with("INFO", "Test message")

def test_logger_helper_methods():
    """Tests helper methods info, error, warning, debug."""
    provider_mock = Mock(spec=LoggerProvider)
    Logger._instance = None  # Reset singleton
    logger = Logger(provider_mock)

    logger.info("Info message")
    provider_mock.log.assert_called_with("INFO", "Info message")

    logger.error("Error message")
    provider_mock.log.assert_called_with("ERROR", "Error message")

    logger.warning("Warning message")
    provider_mock.log.assert_called_with("WARNING", "Warning message")

    logger.debug("Debug message")
    provider_mock.log.assert_called_with("DEBUG", "Debug message")

def test_logger_reset_singleton():
    """Ensures Logger can be reinitialized after resetting the singleton."""
    provider_mock1 = Mock(spec=LoggerProvider)
    provider_mock2 = Mock(spec=LoggerProvider)
    
    Logger._instance = None  # Reset singleton
    logger1 = Logger(provider_mock1)
    Logger._instance = None  # Reset singleton again
    logger2 = Logger(provider_mock2)
    
    assert logger1 is not logger2  # New instance should be created
    assert logger2.provider is provider_mock2

def test_logger_invalid_log_level():
    """Ensures Logger handles invalid log levels gracefully."""
    provider_mock = Mock(spec=LoggerProvider)
    Logger._instance = None  # Reset singleton
    logger = Logger(provider_mock)

    logger.log("INVALID", "Invalid log level message")
    provider_mock.log.assert_called_with("INVALID", "Invalid log level message")


from flotorch_core.embedding.embedding import BaseEmbedding
from typing import List, Dict
from flotorch_core.chunking.chunking import Chunk
from flotorch_core.embedding.embedding import Embeddings, EmbeddingList
from flotorch_core.guardrails.guardrails import BaseGuardRail


class GuardrailsEmbedding(BaseEmbedding):

    def __init__(self, base_embedding: BaseEmbedding, 
                 base_guardrail: BaseGuardRail):
        super().__init__(base_embedding.dimension, base_embedding.normalize)
        self.base_embedding = base_embedding
        self.base_guardrail = base_guardrail

    def _prepare_chunk(self, chunk: Chunk) -> Dict:
        return self.base_embedding._prepare_chunk(chunk)

    """
    Embeds the chunk.
    :param chunk: The chunk to be embedded.
    :return: The embeddings.
    """
    def embed(self, chunk: Chunk) -> Embeddings:
        guardrail_response = self.base_guardrail.apply_guardrail(text=chunk.data)
        if guardrail_response['action'] == 'GUARDRAIL_INTERVENED':
            return None

        return self.base_embedding.embed(chunk)

    """
    Embeds the list of chunks.
    :param chunks: The list of chunks to be embedded.
    :return: The list of embeddings.
    """
    def embed_list(self, chunks: List[Chunk]) -> EmbeddingList:
        embedding_list = EmbeddingList()
        if not isinstance(chunks, list):
            return embedding_list.append(self.embed(chunks))
        for chunk in chunks:
            embedding = self.embed(chunk)
            if not embedding is None:
                embedding_list.append(embedding)
        return embedding_list



import asyncio
from typing import Any, Dict, List, Optional, Union

from flotorch_core.chunking.chunking import Chunk
from flotorch_core.embedding.embedding import BaseEmbedding
from flotorch_core.evaluator.base_evaluator import BaseEvaluator
from flotorch_core.evaluator.evaluation_item import EvaluationItem
from flotorch_core.evaluator.metrics.metrics_keys import MetricKey
from flotorch_core.evaluator.metrics.ragas_metrics.ragas_metrics import RagasEvaluationMetrics

from langchain.embeddings.base import Embeddings
from ragas.embeddings import LangchainEmbeddingsWrapper
from ragas.llms import LangchainLLMWrapper
from langchain_core.language_models.base import LanguageModelLike
from langchain_core.outputs.generation import Generation
from ragas.evaluation import evaluate
from ragas.dataset_schema import SingleTurnSample, EvaluationDataset
from langchain_core.outputs import LLMResult
from itertools import chain

from flotorch_core.inferencer.inferencer import BaseInferencer

class RagasEvaluator(BaseEvaluator):
    """
    Evaluator that uses RAGAS metrics to score RAG-based QA performance.
    Initializes the RagasEvaluator with the given LLM and embedding wrappers.

    Args:
        evaluator_llm: The LLM to be used by RAGAS metrics (wrapped in LangchainLLMWrapper).
        embedding_llm: The embedding model to be used by RAGAS metrics (wrapped in LangchainEmbeddingsWrapper).
        metric_args: Optional configuration for metrics requiring per-instance arguments.

            Example:
            {
                MetricKey.ASPECT_CRITIC: {
                    "maliciousness": {
                        "name": "maliciousness",
                        "definition": "Is the response harmful?"
                    },
                    "bias": {
                        "name": "bias",
                        "definition": "Is the response biased or discriminatory?"
                    }
                }
            }
    """
    def __init__(
        self,
        evaluator_llm: BaseInferencer,
        embedding_llm: BaseEmbedding,
        metric_args: Optional[
            Dict[Union[str, MetricKey], Dict[str, Dict[str, str]]]
        ] = None
    ):
        self.evaluator_llm = evaluator_llm
        self.embedding_llm = embedding_llm
        self.metric_args = metric_args

        class _EmbeddingWrapper(Embeddings):
            def __init__(self, internal_embedding):
                self.internal_embedding = internal_embedding

            def embed_documents(self, texts: list[str]) -> list[list[float]]:
                return [self._embed_text(text) for text in texts]

            def embed_query(self, text: str) -> list[float]:
                return self._embed_text(text)

            def _embed_text(self, text: str) -> list[float]:
                chunk = Chunk(data=text)
                embedding = self.internal_embedding.embed(chunk)
                return embedding.embeddings[0]
            
        class _LLMWrapper(LanguageModelLike):
            def __init__(self, internal_llm: BaseInferencer):
                self.internal_llm = internal_llm

            def invoke(self, prompt: str) -> str:
                """
                This mimics LangChain's ChatOpenAI behavior.
                """
                metadata, response = self.internal_llm.generate_text(user_query=prompt, context=[])
                return response
            
            async def ainvoke(self, prompt: str) -> str:
                """
                Async interface — RAGAS prefers this if available.
                """
                # Run the sync method in an async wrapper
                return await asyncio.to_thread(self.invoke, prompt)
        
            def generate_prompt(self, prompts: List[str], **kwargs: Any,):
                """
                Sync implementation for prompt generation
                """
                responses = []
                for prompt in prompts:
                    metadata, response = self.internal_llm.generate_text(
                        user_query=prompt.text, 
                        context=[]
                    )
                    responses.append(response)
                
                return LLMResult(generations=[[Generation(text=resp)] for resp in responses])

            async def agenerate_prompt(self, prompts: List[str], **kwargs: Any,) -> LLMResult:
                loop = asyncio.get_event_loop()
                futures = []
                
                for prompt in prompts:
                    futures.append(
                        loop.run_in_executor(
                            None, 
                            self.internal_llm.generate_text,
                            prompt.text,
                            []
                        )
                    )
                    
                results = await asyncio.gather(*futures)
                responses = [resp[1] for resp in results]
                
                return LLMResult(generations=[[Generation(text=resp)] for resp in responses])
        
        wrapped_embedding = LangchainEmbeddingsWrapper(_EmbeddingWrapper(self.embedding_llm))
        wrapped_evaluator_llm = LangchainLLMWrapper(_LLMWrapper(self.evaluator_llm))
        
        RagasEvaluationMetrics.initialize_metrics(
            llm=wrapped_evaluator_llm,
            embeddings=wrapped_embedding,
            metric_args=self.metric_args
        )


    def evaluate(
        self,
        data: List[EvaluationItem],
        metrics: Optional[List[MetricKey]] = None
    ) -> Dict[str, Any]:
        # example to fetch metrics, use like this
        if metrics is None:
            metrics = RagasEvaluationMetrics.available_metrics()

        selected_metrics = list(chain.from_iterable(
            RagasEvaluationMetrics.get_metric(m).values() for m in metrics
        ))

        answer_samples = []
        for item in data:
            sample_params = {
                "user_input": item.question,
                "response": item.generated_answer,
                "reference": item.expected_answer,
                "retrieved_contexts": item.context
            }
            answer_samples.append(SingleTurnSample(**sample_params))

        evaluation_dataset = EvaluationDataset(answer_samples)

        result = evaluate(evaluation_dataset, selected_metrics)
        
        return result



import pytest
from flotorch_core.config.config import Config
from flotorch_core.config.config_provider import ConfigProvider

class MockConfigProvider(ConfigProvider):
    def __init__(self, config_dict=None):
        self.config_dict = config_dict or {}
    
    def get(self, key: str, default=None):
        return self.config_dict.get(key, default)

@pytest.fixture
def mock_provider():
    return MockConfigProvider({
        "AWS_REGION": "us-west-2",
        "OPENSEARCH_HOST": "test-host",
        "OPENSEARCH_PORT": "9200",
        "OPENSEARCH_USERNAME": "test-user",
        "OPENSEARCH_PASSWORD": "test-password",
        "OPENSEARCH_INDEX": "test-index",
        "TASK_TOKEN": "test-token",
        "INPUT_DATA": '{"key": "value"}',
        "experiment_table": "test-experiment-table",
        "experiment_question_metrics_table": "test-metrics-table",
        "sagemaker_role_arn": "test-role-arn"
    })

@pytest.fixture
def config(mock_provider):
    return Config(provider=mock_provider)

def test_config_initialization():
    provider = MockConfigProvider()
    config = Config(provider=provider)
    assert config.provider == provider

def test_get_region(config):
    assert config.get_region() == "us-west-2"

def test_get_region_default():
    provider = MockConfigProvider({})
    config = Config(provider=provider)
    assert config.get_region() == "us-east-1"  # Default value

def test_get_region_missing():
    provider = MockConfigProvider({"AWS_REGION": ""})
    config = Config(provider=provider)
    with pytest.raises(ValueError, match="AWS region is not set"):
        config.get_region()

def test_get_opensearch_host(config):
    assert config.get_opensearch_host() == "test-host"

def test_get_opensearch_host_default():
    provider = MockConfigProvider({})
    config = Config(provider=provider)
    assert config.get_opensearch_host() == "localhost"  # Default value

def test_get_opensearch_port(config):
    assert config.get_opensearch_port() == 9200

def test_get_opensearch_port_default():
    provider = MockConfigProvider({})
    config = Config(provider=provider)
    assert config.get_opensearch_port() == 443  # Default value

def test_get_opensearch_username(config):
    assert config.get_opensearch_username() == "test-user"

def test_get_opensearch_username_missing():
    provider = MockConfigProvider({})
    config = Config(provider=provider)
    with pytest.raises(ValueError, match="OpenSearch username is not set"):
        config.get_opensearch_username()

def test_get_opensearch_password(config):
    assert config.get_opensearch_password() == "test-password"

def test_get_opensearch_password_missing():
    provider = MockConfigProvider({})
    config = Config(provider=provider)
    with pytest.raises(ValueError, match="OpenSearch password is not set"):
        config.get_opensearch_password()

def test_get_opensearch_index(config):
    assert config.get_opensearch_index() == "test-index"

def test_get_opensearch_index_missing():
    provider = MockConfigProvider({})
    config = Config(provider=provider)
    with pytest.raises(ValueError, match="OpenSearch index is not set"):
        config.get_opensearch_index()

def test_get_task_token(config):
    assert config.get_task_token() == "test-token"

def test_get_task_token_missing():
    provider = MockConfigProvider({})
    config = Config(provider=provider)
    with pytest.raises(ValueError, match="task token is not set"):
        config.get_task_token()

def test_get_fargate_input_data(config):
    assert config.get_fargate_input_data() == '{"key": "value"}'

def test_get_fargate_input_data_default():
    provider = MockConfigProvider({})
    config = Config(provider=provider)
    with pytest.raises(ValueError, match="input data is not set"):
        config.get_fargate_input_data()

def test_get_experiment_table_name(config):
    assert config.get_experiment_table_name() == "test-experiment-table"

def test_get_experiment_table_name_missing():
    provider = MockConfigProvider({})
    config = Config(provider=provider)
    with pytest.raises(ValueError, match="experiment table name is not set"):
        config.get_experiment_table_name()

def test_get_experiment_question_metrics_table(config):
    assert config.get_experiment_question_metrics_table() == "test-metrics-table"

def test_get_experiment_question_metrics_table_missing():
    provider = MockConfigProvider({})
    config = Config(provider=provider)
    with pytest.raises(ValueError, match="experiment question metrics table name is not set"):
        config.get_experiment_question_metrics_table()

def test_get_sagemaker_arn_role(config):
    assert config.get_sagemaker_arn_role() == "test-role-arn"

def test_get_sagemaker_arn_role_missing():
    provider = MockConfigProvider({})
    config = Config(provider=provider)
    with pytest.raises(ValueError, match="sagemaker arn role is not set"):
        config.get_sagemaker_arn_role()



import pytest
from flotorch_core.chunking.fixedsize_chunking import FixedSizeChunker
from flotorch_core.chunking.chunking import Chunk

@pytest.fixture
def chunker():
    return FixedSizeChunker(chunk_size=100, chunk_overlap=20)

def test_fixed_size_chunker_initialization():
    # Test normal initialization
    chunker = FixedSizeChunker(chunk_size=100, chunk_overlap=20)
    assert chunker.chunk_size == 400  # 100 * tokens_per_charecter (4)
    assert chunker.chunk_overlap == 80  # 20% of chunk_size

def test_invalid_chunk_size():
    # Test initialization with invalid chunk size
    with pytest.raises(ValueError, match="chunk_size must be positive"):
        FixedSizeChunker(chunk_size=0, chunk_overlap=20)
    
    with pytest.raises(ValueError, match="chunk_size must be positive"):
        FixedSizeChunker(chunk_size=-1, chunk_overlap=20)

def test_invalid_chunk_overlap():
    # Test initialization with invalid chunk overlap
    with pytest.raises(ValueError, match="chunk_overlap must be less than chunk_size"):
        FixedSizeChunker(chunk_size=100, chunk_overlap=100)
    
    with pytest.raises(ValueError, match="chunk_overlap must be less than chunk_size"):
        FixedSizeChunker(chunk_size=100, chunk_overlap=150)

def test_chunk_empty_input(chunker):
    # Test chunking empty input
    with pytest.raises(ValueError, match="Input text cannot be empty or None"):
        chunker.chunk("")

def test_chunk_none_input(chunker):
    # Test chunking None input
    with pytest.raises(ValueError, match="Input text cannot be empty or None"):
        chunker.chunk(None)

def test_basic_chunking(chunker):
    # Test basic chunking functionality
    text = "This is a test text that should be split into multiple chunks based on the specified size."
    chunks = chunker.chunk(text)
    
    assert isinstance(chunks, list)
    assert all(isinstance(chunk, Chunk) for chunk in chunks)
    assert len(chunks) > 0

def test_chunk_overlap(chunker):
    # Test that chunks properly overlap
    text = "a" * 1000  # Create a long text
    chunks = chunker.chunk(text)
    
    # Check if consecutive chunks have overlapping content
    for i in range(len(chunks) - 1):
        current_chunk = chunks[i].data
        next_chunk = chunks[i + 1].data
        
        # The end of the current chunk should appear at the start of the next chunk
        assert current_chunk[-80:] == next_chunk[:80]

def test_chunk_size_consistency(chunker):
    # Test that chunks are consistently sized (except possibly the last one)
    text = "This is a test text " * 50  # Create a long text
    chunks = chunker.chunk(text)
    
    # All chunks except the last one should have the same size
    for i in range(len(chunks) - 1):
        assert len(chunks[i].data) <= chunker.chunk_size

def test_chunking_with_different_sizes():
    # Test chunking with different size configurations
    small_chunker = FixedSizeChunker(chunk_size=50, chunk_overlap=10)
    large_chunker = FixedSizeChunker(chunk_size=200, chunk_overlap=30)
    
    text = "This is a test text " * 20
    
    small_chunks = small_chunker.chunk(text)
    large_chunks = large_chunker.chunk(text)
    
    # Smaller chunk size should result in more chunks
    assert len(small_chunks) > len(large_chunks)

def test_chunking_with_special_characters(chunker):
    # Test chunking text with special characters
    text = "Special characters: !@#$%^&*()_+ \n\t" * 10
    chunks = chunker.chunk(text)
    
    assert len(chunks) > 0
    assert all(isinstance(chunk, Chunk) for chunk in chunks)

def test_chunking_with_unicode(chunker):
    # Test chunking text with unicode characters
    text = "Unicode characters: 你好世界 αβγδ 👋🌍" * 10
    chunks = chunker.chunk(text)
    
    assert len(chunks) > 0
    assert all(isinstance(chunk, Chunk) for chunk in chunks)

def test_chunking_single_word():
    # Test chunking a single word that's smaller than chunk size
    chunker = FixedSizeChunker(chunk_size=100, chunk_overlap=20)
    text = "Hello"
    chunks = chunker.chunk(text)
    
    assert len(chunks) == 1
    assert chunks[0].data == "Hello"

def test_chunking_exact_size():
    # Test chunking text that exactly matches chunk size
    chunker = FixedSizeChunker(chunk_size=5, chunk_overlap=0)
    text = "12345"  # Exactly 5 characters
    chunks = chunker.chunk(text)
    
    assert len(chunks) == 1
    assert len(chunks[0].data) == 5

def test_chunking_with_spaces(chunker):
    # Test chunking text with multiple spaces
    text = "This    has    multiple    spaces    between    words"
    chunks = chunker.chunk(text)
    
    assert len(chunks) > 0
    # Verify spaces are normalized to single spaces
    assert chunks[0].data == "This has multiple spaces between words"
    # Verify no multiple spaces are preserved (as per BaseChunker behavior)
    assert "    " not in chunks[0].data

def test_space_normalization(chunker):
    # Test that various types of whitespace are normalized
    text = "This  has\tmultiple\n\rspaces\f\vbetween\t\t\twords"
    chunks = chunker.chunk(text)
    
    assert len(chunks) > 0
    assert chunks[0].data == "This has multiple spaces between words"
    
    # Verify no original whitespace characters remain
    assert "\t" not in chunks[0].data
    assert "\n" not in chunks[0].data
    assert "\r" not in chunks[0].data
    assert "\f" not in chunks[0].data
    assert "\v" not in chunks[0].data
    assert "  " not in chunks[0].data  # No double spaces

def test_chunking_preserves_single_spaces(chunker):
    # Test that single spaces between words are preserved
    text = "This has normal spaces between words"
    chunks = chunker.chunk(text)
    
    assert len(chunks) > 0
    assert chunks[0].data == "This has normal spaces between words"
    # Verify words are still separated by single spaces
    assert " " in chunks[0].data
    # Let's count the actual spaces:
    # "This[1]has[2]normal[3]spaces[4]between[5]words"
    assert chunks[0].data.count(" ") == 5  

def test_chunking_with_leading_trailing_spaces(chunker):
    # Test handling of leading and trailing spaces
    text = "   Leading and trailing spaces   "
    chunks = chunker.chunk(text)
    
    assert len(chunks) > 0
    # Verify leading and trailing spaces are trimmed
    assert chunks[0].data == "Leading and trailing spaces"

def test_chunking_with_newlines(chunker):
    # Test chunking text with newlines
    text = "Line 1\nLine 2\nLine 3\n" * 20
    chunks = chunker.chunk(text)
    
    assert len(chunks) > 0
    # Verify newlines are converted to spaces
    assert "\n" not in chunks[0].data

def test_large_text_performance():
    # Test performance with large text
    chunker = FixedSizeChunker(chunk_size=100, chunk_overlap=20)
    large_text = "Sample text " * 1000
    
    chunks = chunker.chunk(large_text)
    assert len(chunks) > 0

def test_chunk_content_preservation(chunker):
    # Test that the original content is preserved across chunks
    text = "This is a specific test text that should be preserved across chunks."
    chunks = chunker.chunk(text)
    
    # Reconstruct text from chunks (considering overlap)
    reconstructed = chunks[0].data
    for i in range(1, len(chunks)):
        current_chunk = chunks[i].data
        overlap_size = chunker.chunk_overlap
        reconstructed += current_chunk[overlap_size:]
    
    # Remove any extra spaces that might have been added during chunking
    original_cleaned = ' '.join(text.split())
    reconstructed_cleaned = ' '.join(reconstructed.split())
    
    assert original_cleaned == reconstructed_cleaned



from typing import Union, Optional, Mapping, Dict

from ragas.metrics import (
    Faithfulness,
    ResponseRelevancy,
    LLMContextPrecisionWithReference,
    AspectCritic,
)

from flotorch_core.evaluator.metrics.base_metrics import BaseEvaluationMetric
from flotorch_core.evaluator.metrics.metrics_keys import MetricKey


class RagasEvaluationMetrics(BaseEvaluationMetric):
    """
    Registry of RAGAS metric classes and their initialized instances.
    """

    _registry = {
        MetricKey.CONTEXT_PRECISION: {
            "class": LLMContextPrecisionWithReference,
            "requires": ["llm"]
        },
        MetricKey.ASPECT_CRITIC: {
            "class": AspectCritic,
            "requires": ["llm"],
            "metric_args": ["name", "definition"]
        },
        MetricKey.FAITHFULNESS: {
            "class": Faithfulness,
            "requires": ["llm"]
        },
        MetricKey.ANSWER_RELEVANCE: {
            "class": ResponseRelevancy,
            "requires": ["llm", "embeddings"]
        },
    }

    """
    _initialized_metrics: A dictionary like:
        {
            "aspect_critic": {
                "maliciousness": <AspectCritic(...)>,
                "bias": <AspectCritic(...)>
            },
            "faithfulness": {
                "default": <Faithfulness(...)>
            },
            ...
        }
    """
    _initialized_metrics: Dict[str, Dict[str, object]] = {}

    @classmethod
    def available_metrics(cls) -> list[str]:
        return list(cls._initialized_metrics.keys())

    @classmethod
    def initialize_metrics(
        cls,
        llm,
        embeddings,
        metric_args: Optional[
            Mapping[Union[str, MetricKey], Dict[str, Dict[str, str]]]
        ] = None,
    ):
        """
        Initializes metrics based on the registry and stores them in a dict:
        Args:
            llm: A language model wrapper (e.g., LangchainLLMWrapper)
            embeddings: An embedding wrapper (e.g., LangchainEmbeddingsWrapper)
            metric_args (Optional): A nested dictionary specifying arguments for
                per-instance configuration of metrics that require additional input.
                The structure should be:

                {
                    MetricKey.ASPECT_CRITIC: {
                        "maliciousness": {
                            "name": "maliciousness",
                            "definition": "Is the response harmful?"
                        },
                        "bias": {
                            "name": "bias",
                            "definition": "Is the response biased or discriminatory?"
                        }
                    }
                }
        """
        cls._initialized_metrics = {}
        metric_args = metric_args or {}

        for key, config in cls._registry.items():
            base_args = {}
            if "llm" in config["requires"]:
                base_args["llm"] = llm
            if "embeddings" in config["requires"]:
                base_args["embeddings"] = embeddings

            metric_class = config["class"]
            key_str = key.value

            if "metric_args" in config:
                arg_map = metric_args.get(key, {})
                cls._initialized_metrics[key_str] = {}

                for identifier, arg_config in arg_map.items():
                    missing = [param for param in config["metric_args"] if param not in arg_config]
                    if missing:
                        raise ValueError(f"Metric '{key}' is missing required args: {', '.join(missing)}")

                    full_args = base_args | {param: arg_config[param] for param in config["metric_args"]}
                    cls._initialized_metrics[key_str][identifier] = metric_class(**full_args)

            else:
                cls._initialized_metrics[key_str] = {
                    "default": metric_class(**base_args)
                }

    @classmethod
    def get_metric(cls, key: Union[str, MetricKey]) -> Dict[str, object]:
        """
        Returns a dictionary of initialized metric instances for the given metric key.

        Args:
            key (Union[str, MetricKey]): The metric key to fetch (e.g., MetricKey.FAITHFULNESS).

        Returns:
            Dict[str, object]: A dictionary where:
                - Each key is an identifier (e.g., "default", "maliciousness", "bias")
                - Each value is an initialized RAGAS metric instance

        Example:
            {
                "aspect_critic": {
                    "maliciousness": <AspectCritic(...)>,
                    "bias": <AspectCritic(...)>
                },
                "faithfulness": {
                    "default": <Faithfulness(...)>
                }
            }

        Raises:
            ValueError: If the metric key has not been initialized.
        """
        if isinstance(key, MetricKey):
            key = key.value

        if key not in cls._initialized_metrics:
            raise ValueError(f"Metric '{key}' has not been initialized. Did you forget to call `initialize_metrics()`?")
        return cls._initialized_metrics[key]



from abc import ABC, abstractmethod
import boto3

class BaseGuardRail(ABC):

    def __init__(self, prompt=True, response=True):
        self.prompt = prompt
        self.response = response
        
    @abstractmethod
    def apply_guardrail(self, text: str,
        source: str = 'INPUT'):
        pass

class BedrockGuardrail(BaseGuardRail):
    def __init__(self, guardrail_id: str, guardrail_version: str, region_name: str = 'us-east-1', runtime_client = None):
        self.guardrail_id = guardrail_id
        self.guardrail_version = guardrail_version
        self.runtime_client = runtime_client or boto3.client('bedrock-runtime', region_name=region_name)
        
    def apply_guardrail(self, text: str,
        source: str = 'INPUT'):
        try:
            request_params = {
                'guardrailIdentifier': self.guardrail_id,
                'guardrailVersion': self.guardrail_version,
                'source': source,
                'content': [{"text": {"text": text}}]
            }
            response = self.runtime_client.apply_guardrail(**request_params)
            return response
        except Exception as e:
            print(f"Error applying guardrail: {str(e)}")
            raise


from abc import ABC, abstractmethod
from typing import List

class BaseEvaluationMetric(ABC):
    """
    Abstract base class to define metric registries.
    """
    @classmethod
    @abstractmethod
    def available_metrics(cls) -> List[str]:
        """
        Returns a list of available metrics.
        """
        pass

    @classmethod
    @abstractmethod
    def get_metric(cls, key: str):
        """
        Returns the metric associated with the key.
        """
        pass

    @classmethod
    @abstractmethod
    def initialize_metrics(cls, llm, embeddings):
        """
        Initializes the metrics with the provided LLM and embeddings.
        """
        pass


import io
import unittest
from unittest.mock import Mock, patch

import pytest
from PyPDF2 import PdfReader

from flotorch_core.storage.storage import StorageProvider
# Fix the import path to match your project structure
from flotorch_core.reader.pdf_reader import PDFReader


class TestPDFReader(unittest.TestCase):
    def setUp(self):
        # Create a mock storage provider
        self.mock_storage_provider = Mock(spec=StorageProvider)
        
        # Initialize the PDFReader with the mock storage provider
        self.pdf_reader = PDFReader(self.mock_storage_provider)
        
        # Sample PDF content - this would normally be binary PDF data
        # For testing purposes, we'll mock this later
        self.sample_pdf_data = b"mock pdf data"

    @patch('flotorch_core.reader.pdf_reader.PdfReader')
    def test_read_pdf_single_file(self, mock_pdf_reader_class):
        # Configure the mock storage provider to return our sample data
        self.mock_storage_provider.read.return_value = [self.sample_pdf_data]
        
        # Configure the mock PdfReader
        mock_pdf_reader = Mock()
        mock_pdf_reader_class.return_value = mock_pdf_reader
        
        # Create a mock page
        mock_page = Mock()
        mock_page.extract_text.return_value = "This is sample text from a PDF page."
        
        # Set up the reader to return our mock page
        mock_pdf_reader.pages = [mock_page]
        
        # Call the method under test
        result = self.pdf_reader.read_pdf("test/path/to/pdf")
        
        # Assertions
        self.mock_storage_provider.read.assert_called_once_with("test/path/to/pdf")
        mock_pdf_reader_class.assert_called_once()
        self.assertEqual(result, ["This is sample text from a PDF page."])
        
        # Verify the BytesIO was created with our sample data
        # We can check this through the mock_pdf_reader_class call args
        args, _ = mock_pdf_reader_class.call_args
        self.assertIsInstance(args[0], io.BytesIO)
        # Reset the stream position to check content
        args[0].seek(0)
        self.assertEqual(args[0].read(), self.sample_pdf_data)
    
    @patch('flotorch_core.reader.pdf_reader.PdfReader')
    def test_read_pdf_multiple_files(self, mock_pdf_reader_class):
        # Configure the mock storage provider to return multiple files
        self.mock_storage_provider.read.return_value = [self.sample_pdf_data, self.sample_pdf_data]
        
        # Configure the mock PdfReader
        mock_pdf_reader = Mock()
        mock_pdf_reader_class.return_value = mock_pdf_reader
        
        # Create a mock page
        mock_page = Mock()
        mock_page.extract_text.return_value = "This is sample text from a PDF page."
        
        # Set up the reader to return our mock page
        mock_pdf_reader.pages = [mock_page]
        
        # Call the method under test
        result = self.pdf_reader.read_pdf("test/path/to/pdfs")
        
        # Assertions
        self.mock_storage_provider.read.assert_called_once_with("test/path/to/pdfs")
        self.assertEqual(mock_pdf_reader_class.call_count, 2)
        self.assertEqual(result, ["This is sample text from a PDF page.", "This is sample text from a PDF page."])
    
    @patch('flotorch_core.reader.pdf_reader.PdfReader')
    def test_read_pdf_multiple_pages(self, mock_pdf_reader_class):
        # Configure the mock storage provider to return our sample data
        self.mock_storage_provider.read.return_value = [self.sample_pdf_data]
        
        # Configure the mock PdfReader
        mock_pdf_reader = Mock()
        mock_pdf_reader_class.return_value = mock_pdf_reader
        
        # Create mock pages
        mock_page1 = Mock()
        mock_page1.extract_text.return_value = "Page 1 content."
        mock_page2 = Mock()
        mock_page2.extract_text.return_value = "Page 2 content."
        
        # Set up the reader to return our mock pages
        mock_pdf_reader.pages = [mock_page1, mock_page2]
        
        # Call the method under test
        result = self.pdf_reader.read_pdf("test/path/to/pdf")
        
        # Assertions
        self.assertEqual(result, ["Page 1 content.Page 2 content."])
    
    @patch('flotorch_core.reader.pdf_reader.PdfReader')
    def test_read_pdf_empty_page(self, mock_pdf_reader_class):
        # Configure the mock storage provider to return our sample data
        self.mock_storage_provider.read.return_value = [self.sample_pdf_data]
        
        # Configure the mock PdfReader
        mock_pdf_reader = Mock()
        mock_pdf_reader_class.return_value = mock_pdf_reader
        
        # Create mock pages with one empty page
        mock_page1 = Mock()
        mock_page1.extract_text.return_value = "Page content."
        mock_page2 = Mock()
        mock_page2.extract_text.return_value = None
        
        # Set up the reader to return our mock pages
        mock_pdf_reader.pages = [mock_page1, mock_page2]
        
        # Call the method under test
        result = self.pdf_reader.read_pdf("test/path/to/pdf")
        
        # Assertions
        self.assertEqual(result, ["Page content."])
    
    def test_read_pdf_empty_response(self):
        # Configure the mock storage provider to return None
        self.mock_storage_provider.read.return_value = [None]
        
        # Call the method under test
        result = self.pdf_reader.read_pdf("test/path/to/nonexistent")
        
        # Assertions
        self.assertEqual(result, [])
        self.mock_storage_provider.read.assert_called_once_with("test/path/to/nonexistent")

    @patch('flotorch_core.reader.pdf_reader.PdfReader')
    def test_read_pdf_unicode_characters(self, mock_pdf_reader_class):
        # Configure the mock storage provider to return our sample data
        self.mock_storage_provider.read.return_value = [self.sample_pdf_data]
        
        # Configure the mock PdfReader
        mock_pdf_reader = Mock()
        mock_pdf_reader_class.return_value = mock_pdf_reader
        
        # Create a mock page with various Unicode content including all requested examples
        unicode_text = (
            '"question": "¿Qué es la IA?", '
            '"answer": "Inteligencia Artificial 🤖", '
            '"chinese": "人工智能", '
            '"arabic": "الذكاء الاصطناعي", '
            '"hindi": "आर्टिफिशियल इंटेलिजेंस", '
            '"japanese": "人工知能", '
            '"emoji": "🔥🚀"'
        )
        
        mock_page = Mock()
        mock_page.extract_text.return_value = unicode_text
        
        # Set up the reader to return our mock page
        mock_pdf_reader.pages = [mock_page]
        
        # Call the method under test
        result = self.pdf_reader.read_pdf("test/path/to/unicode_pdf")
        
        # Assertions
        self.assertEqual(result, [unicode_text])
        
        # Verify each specific character set is preserved correctly
        self.assertIn("¿Qué es la IA?", result[0])  # Spanish with accents
        self.assertIn("Inteligencia Artificial 🤖", result[0])  # Spanish with robot emoji
        self.assertIn("人工智能", result[0])  # Chinese
        self.assertIn("الذكاء الاصطناعي", result[0])  # Arabic
        self.assertIn("आर्टिफिशियल इंटेलिजेंस", result[0])  # Hindi
        self.assertIn("人工知能", result[0])  # Japanese
        self.assertIn("🔥🚀", result[0])  # Emojis


if __name__ == "__main__":
    unittest.main()


import boto3
from flotorch_core.storage.db.db_storage import DBStorage
from flotorch_core.logger.global_logger import get_logger
from flotorch_core.utils.db_utils import DBUtils
from botocore.exceptions import ClientError
from typing import List, Dict, Any, Optional

logger = get_logger()

class DynamoDB(DBStorage):
    def __init__(self, table_name, region_name='us-east-1'):
        self.table_name = table_name
        self.dynamodb = boto3.resource('dynamodb', region_name=region_name)
        self.table = self.dynamodb.Table(table_name)
        self.primary_key_fields = [key['AttributeName'] for key in self.table.key_schema]

    def write(self, item: dict):
        try:
            serialized_data = DBUtils.serialize_data(item)
            self.table.put_item(Item=serialized_data)
            return True
        except ClientError as e:
            logger.error(f"Error writing to DynamoDB: {e}")
            return False

    def read(self, keys: Optional[Dict[str, Any]] = None) -> List[Dict[str, Any]]:
        items = []
        try:
            if not keys:
                logger.info("No key provided, retrieving all items.")
                response = self.table.scan()
                items.extend(response.get('Items', []))

                # Handle pagination if more results exist
                while 'LastEvaluatedKey' in response:
                    response = self.table.scan(
                        ExclusiveStartKey=response['LastEvaluatedKey']
                    )
                    items.extend(response.get('Items', []))
                return items

            # If keys are provided, retrieved based on it
            if set(keys.keys()) == set(self.primary_key_fields):
                response = self.table.get_item(Key=keys)
                item = response.get('Item', None)
                return [item] if item else []

            # Fallback to scan with filters using pagination
            filter_expression_parts = []
            expression_values = {}
            expression_names = {}

            for k, v in keys.items():
                placeholder_key = f"#k_{k}" 
                placeholder_value = f":v_{k}"
                filter_expression_parts.append(f"{placeholder_key} = {placeholder_value}")
                expression_names[placeholder_key] = k
                expression_values[placeholder_value] = v
            
            filter_expression = " AND ".join(filter_expression_parts)

            response = self.table.scan(
                FilterExpression=filter_expression,
                ExpressionAttributeNames=expression_names,
                ExpressionAttributeValues=expression_values
            )
            items.extend(response.get('Items', []))

            # Handle pagination if more results exist
            while 'LastEvaluatedKey' in response:
                response = self.table.scan(
                    FilterExpression=filter_expression,
                    ExpressionAttributeNames=expression_names,
                    ExpressionAttributeValues=expression_values,
                    ExclusiveStartKey=response['LastEvaluatedKey']
                )
                items.extend(response.get('Items', []))

            return items
        except ClientError as e:
            logger.error(f"Error reading from DynamoDB: {e}")
            return []
        except Exception as e:
            logger.error(f"An unexpected error occurred: {e}")
            return []
    
    def bulk_write(self, items: list):
        with self.table.batch_writer() as batch:
            for item in items:
                batch.put_item(Item=item)
        return True
    
    def update(self, key: Dict[str, Any], data: Dict[str, Any]) -> bool:
        """
        Update method accepts:
        - `key`: Unique identifier to find the record (e.g., {'id': 123})
        - `data`: Fields to be updated with new values (e.g., {'status': 'completed'})
        """
        try:
            # Dynamically construct UpdateExpression and ExpressionAttributeValues
            update_expression = "SET " + ", ".join(f"{k} = :{k}" for k in data.keys())
            expression_values = {f":{k}": v for k, v in data.items()}

            self.table.update_item(
                Key=key,
                UpdateExpression=update_expression,
                ExpressionAttributeValues=expression_values,
                ReturnValues="UPDATED_NEW"
            )
            return True
        except ClientError as e:
            logger.error(f"Error updating DynamoDB: {e}")
            return False
        
    def delete(self, key: Dict[str, Any]) -> bool:
        try:
            response = self.table.delete_item(
                Key=key,
            )
            logger.info(f"Item with key {key} deleted successfully.")
            return True
        except ClientError as e:
            logger.error(f"Error deleting item with key {key} from DynamoDB: {e}")
            return False        


from abc import ABC, abstractmethod
from typing import List, Dict


class BaseReranker(ABC):
    """
    Abstract base class for document rerankers.
    Defines the interface for reranking documents.
    """

    def __init__(self, region: str, rerank_model_id: str):
        """
        Initializes the BaseReranker with AWS region and model ID.

        Args:
            region (str): The AWS region to use.
            rerank_model_id (str): The model ID for reranking.
        """
        self.region = region
        self.rerank_model_id = rerank_model_id

    @abstractmethod
    def rerank_documents(self, input_prompt: str, retrieved_documents: List[Dict[str, str]]) -> List[Dict[str, str]]:
        """
        Reranks a list of documents based on a query.

        Args:
            input_prompt (str): The query for reranking.
            retrieved_documents (List[Dict[str, str]]): List of documents to be reranked.
                                                        Each document is expected to be a dictionary
                                                        with at least a 'text' key.

        Returns:
            List[Dict[str, str]]: A list of reranked documents in order of relevance.
                                  The structure of the returned documents should match the
                                  input documents (e.g., still include 'text' key).
        """
        pass


import json
from unittest import TestCase
import pytest
from unittest.mock import Mock
from flotorch_core.reader.json_reader import JSONReader
from flotorch_core.storage.storage import StorageProvider

# Test model class that accepts any attributes
class TestModel:
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __eq__(self, other):
        if not isinstance(other, TestModel):
            return False
        return self.__dict__ == other.__dict__

class TestJSONReader:
    @pytest.fixture
    def storage_mock(self):
        """Fixture to create a mock storage provider"""
        return Mock(spec=StorageProvider)

    @pytest.fixture
    def json_reader(self, storage_mock):
        """Fixture to create a JSONReader instance with mock storage"""
        return JSONReader(storage_mock)

    def test_init(self, storage_mock):
        """Test JSONReader initialization"""
        reader = JSONReader(storage_mock)
        assert reader.storage_provider == storage_mock

    def test_read_single_object(self, json_reader, storage_mock):
        """Test reading a single JSON object"""
        test_json = '{"field1": "value1", "field2": "value2"}'
        storage_mock.read.return_value = [test_json.encode('utf-8')]

        result = json_reader.read("test.json")

        assert isinstance(result, dict)
        assert result == {"field1": "value1", "field2": "value2"}

    def test_read_array(self, json_reader, storage_mock):
        """Test reading a JSON array"""
        test_json = '[{"field1": "v1", "field2": "v2"}, {"field1": "v3", "field2": "v4"}]'
        storage_mock.read.return_value = [test_json.encode('utf-8')]

        result = json_reader.read("test.json")

        assert isinstance(result, list)
        assert len(result) == 2
        assert result[0] == {"field1": "v1", "field2": "v2"}
        assert result[1] == {"field1": "v3", "field2": "v4"}

    def test_read_chunked_data(self, json_reader, storage_mock):
        """Test reading JSON data that comes in chunks"""
        storage_mock.read.return_value = [
            b'{"field1": "',
            b'partial value", "field2": "another value"}'
        ]

        result = json_reader.read("test.json")

        assert isinstance(result, dict)
        assert result == {"field1": "partial value", "field2": "another value"}

    def test_read_invalid_json(self, json_reader, storage_mock):
        """Test reading invalid JSON data"""
        storage_mock.read.return_value = [b'{"invalid": json}']

        with pytest.raises(json.JSONDecodeError):
            json_reader.read("test.json")

    def test_read_empty_json(self, json_reader, storage_mock):
        """Test handling empty JSON data"""
        storage_mock.read.return_value = [b'']

        with pytest.raises(json.JSONDecodeError):
            json_reader.read("test.json")

    def test_read_with_unicode(self, json_reader, storage_mock):
        """Test reading JSON with unicode characters from multiple languages"""
        test_json = '''
        {
            "spanish": "¿Qué es la IA?", 
            "spanish_answer": "Inteligencia Artificial 🤖",
            "chinese": "人工智能",
            "arabic": "الذكاء الاصطناعي",
            "hindi": "आर्टिफिशियल इंटेलिजेंस",
            "japanese": "人工知能",
            "emoji": "🔥🚀"
        }
        '''
        storage_mock.read.return_value = [test_json.encode('utf-8')]

        result = json_reader.read("test.json")

        assert result["spanish"] == "¿Qué es la IA?"
        assert result["spanish_answer"] == "Inteligencia Artificial 🤖"
        assert result["chinese"] == "人工智能"  # Chinese
        assert result["arabic"] == "الذكاء الاصطناعي"  # Arabic
        assert result["hindi"] == "आर्टिफिशियल इंटेलिजेंस"  # Hindi
        assert result["japanese"] == "人工知能"  # Japanese
        assert result["emoji"] == "🔥🚀"  # Emojis

    def test_storage_provider_error(self, json_reader, storage_mock):
        """Test handling storage provider errors"""
        storage_mock.read.side_effect = Exception("Storage error")

        with pytest.raises(Exception, match="Storage error"):
            json_reader.read("test.json")

    def test_read_as_model_single_object(self, json_reader, storage_mock):
        """Test converting single JSON object to TestModel"""
        test_json = '{"field1": "value1", "field2": "value2"}'
        storage_mock.read.return_value = [test_json.encode('utf-8')]

        result = json_reader.read_as_model("test.json", TestModel)

        assert isinstance(result, list)
        assert len(result) == 1
        assert isinstance(result[0], TestModel)
        assert result[0].field1 == "value1"
        assert result[0].field2 == "value2"

    def test_read_as_model_array(self, json_reader, storage_mock):
        """Test converting JSON array to TestModel objects"""
        test_json = '[{"field1": "v1", "field2": "v2"}, {"field1": "v3", "field2": "v4"}]'
        storage_mock.read.return_value = [test_json.encode('utf-8')]

        result = json_reader.read_as_model("test.json", TestModel)

        assert isinstance(result, list)
        assert len(result) == 2
        assert all(isinstance(item, TestModel) for item in result)
        assert result[0].field1 == "v1"
        assert result[0].field2 == "v2"
        assert result[1].field1 == "v3"
        assert result[1].field2 == "v4"

    def test_read_as_model_different_field_names(self, json_reader, storage_mock):
        """Test handling JSON with different field names across objects"""
        test_json = '[{"field1": "v1", "field2": "v2"}, {"field3": "v3", "field4": "v4"}]'
        storage_mock.read.return_value = [test_json.encode('utf-8')]

        result = json_reader.read_as_model("test.json", TestModel)

        assert isinstance(result, list)
        assert len(result) == 2
        assert result[0].field1 == "v1"
        assert result[0].field2 == "v2"
        assert result[1].field3 == "v3"
        assert result[1].field4 == "v4"

    def test_read_as_model_mixed_types(self, json_reader, storage_mock):
        """Test handling JSON with various data types"""
        test_json = '''
        {
            "string_field": "text value",
            "number_field": 42,
            "float_field": 3.14,
            "boolean_field": true,
            "null_field": null,
            "array_field": [1, 2, 3],
            "object_field": {"nested": "value"}
        }
        '''
        storage_mock.read.return_value = [test_json.encode('utf-8')]

        result = json_reader.read_as_model("test.json", TestModel)

        assert isinstance(result, list)
        assert len(result) == 1
        assert result[0].string_field == "text value"
        assert result[0].number_field == 42
        assert result[0].float_field == 3.14
        assert result[0].boolean_field is True
        assert result[0].null_field is None
        assert result[0].array_field == [1, 2, 3]
        assert result[0].object_field == {"nested": "value"}

    def test_read_as_model_custom_validation(self, json_reader, storage_mock):
        """Test with a model that has custom validation"""
        class ModelWithValidation:
            def __init__(self, **kwargs):
                # Only require at least one field to be present
                if not kwargs:
                    raise ValueError("At least one field is required")
                for key, value in kwargs.items():
                    setattr(self, key, value)

        test_json = '[{"valid": "data"}, {}]'  # Second object is empty
        storage_mock.read.return_value = [test_json.encode('utf-8')]

        with pytest.raises(ValueError):
            json_reader.read_as_model("test.json", ModelWithValidation)

    def test_read_as_model_with_nested_objects(self, json_reader, storage_mock):
        """Test handling JSON with nested objects"""
        test_json = '''
        {
            "main_field": "value",
            "nested": {
                "sub_field1": "nested value",
                "sub_field2": 42
            }
        }
        '''
        storage_mock.read.return_value = [test_json.encode('utf-8')]

        result = json_reader.read_as_model("test.json", TestModel)

        assert isinstance(result, list)
        assert len(result) == 1
        assert result[0].main_field == "value"
        assert result[0].nested == {"sub_field1": "nested value", "sub_field2": 42}


import pytest
from unittest.mock import Mock, patch
from flotorch_core.rerank.rerank import BedrockReranker

@pytest.fixture
def mock_bedrock_client():
    """Creates a mock Bedrock client."""
    return Mock()

def test_bedrock_reranker_initialization_with_mock_client(mock_bedrock_client):
    """Tests initialization when a mock Bedrock client is provided."""
    reranker = BedrockReranker(region="us-west-2", rerank_model_id="test-model", bedrock_client=mock_bedrock_client)
    assert reranker.region == "us-west-2"
    assert reranker.rerank_model_id == "test-model"
    assert reranker.bedrock_agent_runtime == mock_bedrock_client

@patch("flotorch_core.rerank.rerank.boto3.client")
def test_bedrock_reranker_initialization_without_mock_client(mock_boto3_client):
    """Tests initialization when no client is provided (boto3 should be called)."""
    mock_boto3_client.return_value = Mock()  # Mock boto3 client return
    reranker = BedrockReranker(region="us-west-2", rerank_model_id="test-model")

    mock_boto3_client.assert_called_once_with("bedrock-agent-runtime", region_name="us-west-2")
    assert reranker.rerank_model_id == "test-model"

def test_rerank_documents_empty_list(mock_bedrock_client):
    """Tests reranking with an empty document list (should return empty)."""
    reranker = BedrockReranker(region="us-west-2", rerank_model_id="test-model", bedrock_client=mock_bedrock_client)
    result = reranker.rerank_documents("query", [])
    assert result == []

def test_rerank_documents_success(mock_bedrock_client):
    """Tests reranking with valid inputs and mocked API response."""
    reranker = BedrockReranker(region="us-west-2", rerank_model_id="test-model", bedrock_client=mock_bedrock_client)

    input_prompt = "Query"
    retrieved_documents = [{"text": "Doc1"}, {"text": "Doc2"}, {"text": "Doc3"}]

    # Mock response from Bedrock API
    mock_bedrock_client.rerank.return_value = {
        "results": [{"index": 2}, {"index": 0}, {"index": 1}]
    }

    result = reranker.rerank_documents(input_prompt, retrieved_documents)

    expected_result = [{"text": "Doc3"}, {"text": "Doc1"}, {"text": "Doc2"}]
    assert result == expected_result
    mock_bedrock_client.rerank.assert_called_once()

def test_rerank_documents_no_results(mock_bedrock_client):
    """Tests when Bedrock returns no results (should return empty)."""
    reranker = BedrockReranker(region="us-west-2", rerank_model_id="test-model", bedrock_client=mock_bedrock_client)
    mock_bedrock_client.rerank.return_value = {"results": []}

    result = reranker.rerank_documents("query", [{"text": "Doc1"}])
    assert result == []

def test_rerank_documents_exception_handling(mock_bedrock_client):
    """Tests error handling when Bedrock API raises an exception."""
    reranker = BedrockReranker(region="us-west-2", rerank_model_id="test-model", bedrock_client=mock_bedrock_client)
    mock_bedrock_client.rerank.side_effect = Exception("Bedrock API error")

    result = reranker.rerank_documents("query", [{"text": "Doc1"}])
    assert result == []

def test_rerank_documents_invalid_structure(mock_bedrock_client):
    """Tests reranking with invalid document structure."""
    reranker = BedrockReranker(region="us-west-2", rerank_model_id="test-model", bedrock_client=mock_bedrock_client)
    invalid_documents = [{"invalid_key": "value"}]  # Missing "text" key

    result = reranker.rerank_documents("query", invalid_documents)
    assert result == []

@patch("flotorch_core.rerank.rerank.boto3.client")
def test_rerank_documents_invalid_api_response(mock_boto3_client):
    """Tests reranking when API response is invalid."""
    mock_boto3_client.return_value.rerank.return_value = {"invalid_key": "value"}  # Missing "results"
    reranker = BedrockReranker(region="us-west-2", rerank_model_id="test-model", bedrock_client=mock_boto3_client)

    result = reranker.rerank_documents("query", [{"text": "Doc1"}])
    assert result == []


from abc import ABC, abstractmethod
import re
from typing import List, Dict

from flotorch_core.chunking.chunking import Chunk

"""
This class is responsible for embedding the text using the Llama model.
"""
class EmbeddingMetadata:
    """
    Initializes the EmbeddingMetadata class.
    :param input_tokens: The number of input tokens.
    :param latency_ms: The latency in milliseconds.
    """
    def __init__(self, input_tokens: int, latency_ms: int):
        self.input_tokens = input_tokens
        self.latency_ms = latency_ms
    
    def append(self, metadata: 'EmbeddingMetadata'):
        self.input_tokens += int(metadata.input_tokens)
        self.latency_ms += int(metadata.latency_ms)

    def to_json(self):
        return {
            'input_token': self.input_tokens,
            'latency_ms': self.latency_ms
        }


class Embeddings:
    """
    Initializes the Embeddings class.
    :param embeddings: The embeddings.
    :param metadata: The metadata.
    """
    def __init__(self, embeddings: List[List[float]], metadata: EmbeddingMetadata, text: str):
        self.embeddings = embeddings
        self.metadata = metadata
        self.text = text
        self.id = ''

    def clean_text_for_vector_db(self, text):
        """
        Cleans the input text by removing quotes, special symbols, extra whitespaces,
        newline (\n), and tab (\t) characters.

        Args:
            text (str): The input text to clean.

        Returns:
            str: The cleaned text.
        """
        # Remove single and double quotes
        text = text.replace('"', '').replace("'", "")
        # Remove special symbols (keeping alphanumerics and spaces)
        text = re.sub(r'[^a-zA-Z0-9\s]', '', text)
        # Remove newlines and tabs
        text = text.replace('\n', ' ').replace('\t', ' ')
        # Normalize whitespace
        text = re.sub(r'\s+', ' ', text)
        # Strip leading and trailing spaces
        return text.strip()

    def to_json(self) -> Dict:
        return {
            "vectors": self.embeddings,
            "text": self.clean_text_for_vector_db(self.text),
            "metadata": {
                    "inputTokens": self.metadata.input_tokens,
                    "latencyMs": self.metadata.latency_ms
                }
        }

class EmbeddingList:
    def __init__(self):
        self.embeddings: List[Embeddings] = []
        self.metadata = EmbeddingMetadata(0, 0)

    def append(self, embeddings: Embeddings):
        self.embeddings.append(embeddings)
        self.metadata.append(embeddings.metadata)

"""
This class is responsible for embedding the text."""
class BaseEmbedding(ABC):
    """
    Initializes the BaseEmbedding class.
    :param dimensions: The dimensions of the embedding.
    :param normalize: Normalize the embedding.
    """

    def __init__(self,  model_id: str, region: str, dimensions: int = 256, normalize: bool = True) -> None:
        super().__init__()
        self.model_id = model_id
        self.region = region
        self.dimension = dimensions
        self.normalize = normalize

    """
    Prepares the chunk for embedding.
    :param chunk: The chunk to be embedded.
    :return: The prepared chunk.
    """
    @abstractmethod
    def _prepare_chunk(self, chunk: Chunk) -> Dict:
        pass

    """
    Embeds the chunk.
    :param chunk: The chunk to be embedded.
    :return: The embeddings.
    """
    @abstractmethod
    def embed(self, chunk: Chunk) -> Embeddings:
        pass

    """
    Embeds the list of chunks.
    :param chunks: The list of chunks to be embedded.
    :return: The list of embeddings.
    """
    def embed_list(self, chunks: List[Chunk]) -> EmbeddingList:
        embedding_list = EmbeddingList()
        if not isinstance(chunks, list):
            return embedding_list.append(self.embed(chunks))
        for chunk in chunks:
            if chunk.child_data:
                for child_chunk in chunk.child_data:
                    embedding = self.embed(child_chunk)
                    embedding.id = chunk.id
                    embedding.text = chunk.data
                    embedding_list.append(embedding)
            else:
                embedding = self.embed(chunk)
                embedding.id = chunk.id
                embedding_list.append(embedding)
        return embedding_list


import os
import pytest
from flotorch_core.config.env_config_provider import EnvConfigProvider

@pytest.fixture
def env_config():
    return EnvConfigProvider()

def test_get_existing_env_variable(env_config):
    # Setup
    test_key = "TEST_VAR"
    test_value = "test_value"
    os.environ[test_key] = test_value
    
    # Execute
    result = env_config.get(test_key)
    
    # Assert
    assert result == test_value
    
    # Cleanup
    del os.environ[test_key]

def test_get_non_existing_env_variable_with_default(env_config):
    # Execute
    result = env_config.get("NON_EXISTING_VAR", default="default_value")
    
    # Assert
    assert result == "default_value"

def test_get_non_existing_env_variable_without_default(env_config):
    # Execute
    result = env_config.get("NON_EXISTING_VAR")
    
    # Assert
    assert result is None



import pytest
import boto3
from unittest.mock import Mock, patch
from flotorch_core.storage.storage_provider_factory import StorageProviderFactory
from flotorch_core.storage.local_storage import LocalStorageProvider
from flotorch_core.storage.s3_storage import S3StorageProvider
from unittest.mock import patch, Mock


def test_create_storage_provider_local_path():
    """Test creating storage provider with local file path"""
    # Test data
    local_path = "file:///Users/dheeratallapragada/Downloads/medical_20_qa_small_questions.json"
    
    # Execute
    storage = StorageProviderFactory.create_storage_provider(local_path)
    
    # Assert
    assert isinstance(storage, LocalStorageProvider)
    assert storage.get_path(local_path) == "/Users/dheeratallapragada/Downloads/medical_20_qa_small_questions.json"

def test_create_storage_provider_s3_path():
    """Test creating storage provider with S3 path"""
    # Test data
    s3_path = "s3://flotorch-data-refact/bfa24db3-ab76-47e7-9ff3-9e942a0a2a63/gt_data/gt.json"
    
    # Execute
    storage = StorageProviderFactory.create_storage_provider(s3_path)
    
    # Assert
    assert isinstance(storage, S3StorageProvider)
    assert storage.get_path(s3_path) == "bfa24db3-ab76-47e7-9ff3-9e942a0a2a63/gt_data/gt.json"

def test_create_storage_provider_invalid_path():
    """Test creating storage provider with invalid path"""
    # Test data
    invalid_path = "invalid://path/to/file.json"
    
    # Assert
    with pytest.raises(ValueError) as exc_info:
        StorageProviderFactory.create_storage_provider(invalid_path)
    assert "Unsupported storage scheme: invalid" in str(exc_info.value)

@patch('flotorch_core.storage.s3_storage.boto3.client', autospec=True)
def test_create_storage_provider_s3_with_mock(mock_client):
    """Test creating storage provider with S3 path using mocked boto3"""
    # Test data
    s3_path = "s3://flotorch-data-refact/test/file.json"

    # Create a fresh mock for the S3 client
    mock_s3_client = Mock()
    mock_client.return_value = mock_s3_client

    # Force reload of the module to ensure our mock is used
    import importlib
    import flotorch_core.storage.s3_storage
    importlib.reload(flotorch_core.storage.s3_storage)

    # Execute
    storage = StorageProviderFactory.create_storage_provider(s3_path)

    # Assert
    assert isinstance(storage, S3StorageProvider)
    assert storage.bucket == "flotorch-data-refact"
    assert storage.get_path(s3_path) == "test/file.json"
    
    # Verify the mock was called
    mock_client.assert_called_with('s3')


# Use below if s3_storage.py accepts s3_client = None
# @patch('boto3.client')
# def test_create_storage_provider_s3_with_mock(mock_boto3_client):
#     """Test creating storage provider with S3 path using mocked boto3"""
#     # Test data
#     s3_path = "s3://flotorch-data-refact/test/file.json"

#     # Mock boto3 client
#     mock_s3_client = Mock()
#     mock_boto3_client.return_value = mock_s3_client

#     # Execute
#     storage = StorageProviderFactory.create_storage_provider(s3_path)

#     # Assert
#     assert isinstance(storage, S3StorageProvider)
#     assert storage.bucket == "flotorch-data-refact"
#     assert storage.get_path(s3_path) == "test/file.json"

#     # Verify that boto3.client('s3') was called once
#     mock_boto3_client.assert_called_once_with('s3')






import logging
from abc import ABC, abstractmethod
from typing import Generator

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class StorageProvider(ABC):
    """
    Abstract base class for storage
    """

    @abstractmethod
    def get_path(self, uri: str) -> str:
        """
        Gets the path from the URI.
        Args:
            uri (str): The URI to get the path from.
        Returns:
            str: The path extracted from the URI.
        """
        pass

    @abstractmethod
    def write(self, path: str, data: bytes) -> None:
        """
        Writes data to the specified path.
        Args:
            path (str): The path to write the data to.
            data (bytes): The data to write.
        """
        pass

    @abstractmethod
    def read(self, path: str) -> Generator[bytes, None, None]:
        """
        Reads data from the specified path.
        Args:
            path (str): The path to read the data from.
        Returns:
            Generator[bytes, None, None]: A generator that yields the data read.
        """
        pass

    def read_as_string(self, path: str) -> Generator[str, None, None]:
        """
        Reads data from the specified path and yields it as a string.
        Args:
            path (str): The path to read the data from.
        Returns:
            Generator[str, None, None]: A generator that yields the data read as a string.
        """
        yield from (data.decode('utf-8') for data in self.read(path))



import pytest
from flotorch_core.chunking.chunking import Chunk, BaseChunker
from typing import List


def test_chunk_initialization():
    # Test basic chunk creation
    data = "test data"
    chunk = Chunk(data)

    assert chunk.data == data
    assert chunk.child_data is None
    assert isinstance(chunk.id, str)  # Verify UUID was created

def test_chunk_add_child():
    # Test adding child data to chunk
    chunk = Chunk("parent data")
    child_data = "child data"
    
    chunk.add_child(child_data)
    
    assert chunk.child_data == [child_data]
    
    # Test adding multiple children
    second_child = "second child"
    chunk.add_child(second_child)
    assert chunk.child_data == [child_data, second_child]

def test_chunk_str_representation():
    # Test string representation of chunk
    chunk = Chunk("parent")
    assert str(chunk) == "Parent Chunk: parent, Chunk: None"
    
    chunk.add_child("child")
    assert str(chunk) == "Parent Chunk: parent, Chunk: ['child']"

# Create a concrete implementation of BaseChunker for testing
class TestChunker(BaseChunker):
    def chunk(self, data: str) -> List[Chunk]:
        cleaned_data = self._clean_data(data)
        return [Chunk(cleaned_data)]

@pytest.fixture
def chunker():
    return TestChunker()

def test_base_chunker_initialization(chunker):
    assert chunker.space == ' '
    assert chunker.separators == [' ', '\t', '\n', '\r', '\f', '\v']
    assert chunker.tokens_per_charecter == 4

def test_clean_data(chunker):
    # Test cleaning data with different separators
    input_text = "Hello\tWorld\nNew\rLine\fTest\vText"
    expected = "Hello World New Line Test Text"
    
    cleaned = chunker._clean_data(input_text)
    assert cleaned == expected

def test_chunk_list(chunker):
    # Test chunking a list of texts
    input_texts = ["Hello World", "Test Text"]
    chunks = chunker.chunk_list(input_texts)
    
    assert len(chunks) == 2
    assert all(isinstance(c, Chunk) for c in chunks)
    assert chunks[0].data == "Hello World"
    assert chunks[1].data == "Test Text"

def test_base_chunker_is_abstract():
    # Verify that BaseChunker cannot be instantiated directly
    with pytest.raises(TypeError):
        BaseChunker()

def test_chunker_with_empty_input(chunker):
    # Test handling empty input
    chunks = chunker.chunk("")
    assert len(chunks) == 1
    assert chunks[0].data == ""

def test_chunker_with_none_input(chunker):
    # Test handling None input
    with pytest.raises(AttributeError):
        chunker.chunk(None)

def test_chunk_list_with_empty_list(chunker):
    # Test chunking empty list
    chunks = chunker.chunk_list([])
    assert chunks == []

def test_chunk_list_with_none_elements(chunker):
    # Test chunking list with None elements
    with pytest.raises(AttributeError):
        chunker.chunk_list([None])

def test_clean_data_with_multiple_spaces(chunker):
    # Test cleaning data with multiple consecutive spaces
    input_text = "Hello    World  Test"
    cleaned = chunker._clean_data(input_text)
    assert cleaned == "Hello    World  Test"  # Should preserve multiple spaces

def test_clean_data_with_mixed_separators(chunker):
    # Test cleaning data with mixed separators
    input_text = "Hello\t\nWorld\r\f\vTest"
    cleaned = chunker._clean_data(input_text)
    assert cleaned == "Hello  World   Test"

# Test edge cases
def test_chunk_with_special_characters(chunker):
    # Test chunking text with special characters
    input_text = "Hello! @#$%^&*()"
    chunks = chunker.chunk(input_text)
    assert chunks[0].data == "Hello! @#$%^&*()"

def test_chunk_with_unicode_characters(chunker):
    # Test chunking text with unicode characters
    input_text = "Hello 世界"
    chunks = chunker.chunk(input_text)
    assert chunks[0].data == "Hello 世界"

def test_chunk_list_with_mixed_content(chunker):
    # Test chunking list with mixed content types
    input_texts = ["Hello", "World\tTest", "Line\nBreak"]
    chunks = chunker.chunk_list(input_texts)
    
    assert len(chunks) == 3
    assert chunks[0].data == "Hello"
    assert chunks[1].data == "World Test"
    assert chunks[2].data == "Line Break"

# Performance test for large inputs
def test_large_input_performance(chunker):
    # Test performance with large input
    large_text = "word " * 1000
    chunks = chunker.chunk(large_text)
    assert len(chunks) > 0

# Test chunk ID uniqueness
def test_chunk_id_uniqueness():
    # Test that each chunk gets a unique ID
    chunk1 = Chunk("data1")
    chunk2 = Chunk("data2")
    assert chunk1.id != chunk2.id



import psycopg2
from psycopg2.extras import RealDictCursor
from flotorch_core.storage.db.db_storage import DBStorage
from flotorch_core.utils.db_utils import DBUtils
from typing import List, Dict, Any, Optional
from flotorch_core.logger.global_logger import get_logger

logger = get_logger()

class PostgresDB(DBStorage):
    def __init__(self, dbname: str, user: str, password: str, table_name: str, host: str = "localhost", port: int = 5432):
        self.dbname = dbname
        self.user = user
        self.password = password
        self.table = f'"{table_name}"'
        self.host = host
        self.port = port
        self.conn = self._connect()

    def _connect(self):
        try:
            return psycopg2.connect(
                dbname=self.dbname,
                user=self.user,
                password=self.password,
                host=self.host,
                port=self.port
            )
        except psycopg2.Error as e:
            logger.error(f"Error connecting to PostgreSQL: {e}")
            return None

    def write(self, item: dict):
        """
        Insert a single item into the specified table.
        """
        if not self.conn:
            return False
        
        item = DBUtils.prepare_item_for_write(item)
        columns = ", ".join([f'"{k}"' for k in item.keys()])
        values = ", ".join([f"%({k})s" for k in item.keys()])
        query = f"INSERT INTO {self.table} ({columns}) VALUES ({values}) ON CONFLICT DO NOTHING"
        
        try:
            with self.conn.cursor() as cur:
                cur.execute(query, item)
                self.conn.commit()
            return True
        except psycopg2.Error as e:
            print(f"Error writing to PostgreSQL: {e}")
            return False

    def read(self, key: Optional[Dict[str, Any]] = None) -> Optional[List[dict]]:
            """
            Retrieve item(s) based on key.
            If no key is provided, retrieves all items from the table.
            """
            if not self.conn:
                logger.error("No database connection available.")
                return None

            try:
                with self.conn.cursor(cursor_factory=RealDictCursor) as cur:
                    if key and len(key) > 0:
                        key_column, key_value = next(iter(key.items()))
                        quoted_key_column = f'"{key_column}"'
                        query = f"SELECT * FROM {self.table} WHERE {quoted_key_column} = %s"
                        cur.execute(query, (key_value,))
                    else:
                        logger.info("No key provided, retrieving all items.")
                        query = f"SELECT * FROM {self.table}"
                        cur.execute(query)
                    
                    result = cur.fetchall()
                    return result if result else []

            except psycopg2.Error as e:
                logger.error(f"Error reading from PostgreSQL table {self.table}: {e}")
                return None
            except Exception as e:
                logger.error(f"An unexpected error occurred while reading from {self.table}: {e}")
                return None

    def bulk_write(self, items: List[dict]):
        """
        Insert multiple items using batch execution.
        """
        if not self.conn or not items:
            return False
        columns = ", ".join(items[0].keys())
        values = ", ".join([f"%({k})s" for k in items[0].keys()])
        query = f"INSERT INTO {self.table} ({columns}) VALUES ({values}) ON CONFLICT DO NOTHING"

        try:
            with self.conn.cursor() as cur:
                cur.executemany(query, items)
                self.conn.commit()
            return True
        except psycopg2.Error as e:
            logger.error(f"Error writing multiple records to PostgreSQL: {e}")
            return False

    def update(self, key: Dict[str, Any], data: Dict[str, Any]) -> bool:
        """
        Update existing record(s) based on the key.
        """
        if not self.conn:
            return False
        key_column, key_value = next(iter(key.items()))
        set_clause = ", ".join([f"{k} = %s" for k in data.keys()])
        query = f"UPDATE {self.table} SET {set_clause} WHERE {key_column} = %s"

        try:
            with self.conn.cursor() as cur:
                cur.execute(query, list(data.values()) + [key_value])
                self.conn.commit()
            return True
        except psycopg2.Error as e:
            logger.error(f"Error updating PostgreSQL: {e}")
            return False

    def close(self):
        """ Close the database connection. """
        if self.conn:
            self.conn.close()
            
    def delete(self, key: Dict[str, Any]) -> bool:
        """
        Delete existing record(s) based on the key_conditions.

        """
        if not self.conn or not key:
            logger.error("Delete: No connection or no key conditions specified.")
            return False

        where_clauses = []
        values = []
        for col, val in key.items():
            where_clauses.append(f'"{col}" = %s')
            values.append(val)
        
        where_expression = " AND ".join(where_clauses)
        query = f"DELETE FROM {self.table} WHERE {where_expression}"

        try:
            with self.conn.cursor() as cur:
                cur.execute(query, tuple(values))
                if cur.rowcount == 0:
                    logger.info(f"Warning: Delete affected 0 rows for key conditions {key}. Item(s) might not exist.")
                else:
                    logger.info(f"Successfully deleted {cur.rowcount} row(s) matching {key}.")
                self.conn.commit()
            return True
        except psycopg2.Error as e:
            logger.error(f"Error deleting from PostgreSQL table {self.table}: {e}")
            self.conn.rollback()
            return False


from abc import ABC, abstractmethod
from typing import List, Dict, Any, Optional
from flotorch_core.evaluator.evaluation_item import EvaluationItem
from flotorch_core.evaluator.metrics.metrics_keys import MetricKey

class BaseEvaluator(ABC):
    """
    Abstract base class for evaluation modules.
    """

    @abstractmethod
    def evaluate(
        self,
        data: List[EvaluationItem],
        metrics: Optional[List[MetricKey]] = None
    ) -> Dict[str, Any]:
        """
        Evaluate the model output against the expected answers.

        Args:
            data (List[EvaluationItem]): List of evaluation inputs.

        Returns:
            Dict[str, Any]: Dictionary of evaluation results.
        """
        pass



from dataclasses import dataclass, field
from typing import List


@dataclass
class EvaluationItem:
    """
    Represents one unit of evaluation — a question and its corresponding model output and metadata.
    """
    question: str
    generated_answer: str
    expected_answer: str
    context: List[str] = field(default_factory=list)



import uuid
from abc import ABC, abstractmethod
from typing import List


class Chunk:
    """
    A class to represent a chunk of text.
    """

    def __init__(self, data):
        """
        Constructs a new Chunk object.
        Args:
            data (str): The data of the chunk.
        """
        self.id = str(uuid.uuid4())
        self.data = data
        self.child_data = None

    def add_child(self, child_data):
        """
        Adds a child to the chunk.
        Args:
            child_data (str): The data of the child.
        """
        if not self.child_data:
            self.child_data = []
        self.child_data.append(child_data)

    def __str__(self):
        """
        Returns a string representation of the Chunk object.
        """
        return f"Parent Chunk: {self.data}, Chunk: {self.child_data}"


class BaseChunker(ABC):
    """
    Abstract base class for chunking text.
    """

    def __init__(self):
        """
        Constructs a new BaseChunker object.
        """
        super().__init__()
        self.space = ' '
        self.separators = [' ', '\t', '\n', '\r', '\f', '\v']
        self.tokens_per_charecter = 4

    @abstractmethod
    def chunk(self, data: str) -> List[Chunk]:
        """
        Chunks the input text.
        Args:
            data (str): The input text to chunk.
        Returns:
            List[Chunk]: A list of Chunk objects.  
        """
        pass

    def chunk_list(self, data: List[str]) -> List[Chunk]:
        """
        Chunks a list of input texts.
        Args:
            data (List[str]): The list of input texts to chunk.
        Returns:
            List[Chunk]: A list of Chunk objects.  
        """
        result = []
        for d in data:
            result.extend(self.chunk(d))
        return result

    def _clean_data(self, data: str) -> str:
        """
        Cleans the input text.
        Args:
            data (str): The input text to clean.
        Returns:
            str: The cleaned text.
        """
        for sep in self.separators:
            data = data.replace(sep, self.space)
        return data



import logging
import boto3
from typing import List, Dict, Optional

from flotorch_core.logger.global_logger import get_logger
from flotorch_core.rerank.reranker import BaseReranker

logger = get_logger()

class BedrockReranker(BaseReranker):
    def __init__(self, region: str, rerank_model_id: str, bedrock_client: Optional[boto3.client] = None):
        """
        Initializes the DocumentReranker with AWS region, model ID, and an optional Bedrock client.

        Args:
            region (str): The AWS region to use.
            rerank_model_id (str): The model ID for reranking.
            bedrock_client (boto3.client, optional): Pre-initialized Bedrock agent runtime client.
        """
        super().__init__(region, rerank_model_id)
        self.bedrock_agent_runtime = bedrock_client or boto3.client('bedrock-agent-runtime', region_name=region)

    def rerank_documents(self, input_prompt: str, retrieved_documents: List[Dict[str, str]]) -> List[Dict[str, str]]:
        """
        Reranks a list of documents based on a query using Amazon Bedrock's reranking model.

        Args:
            input_prompt (str): The query for reranking.
            retrieved_documents (List[Dict[str, str]]): List of documents to be reranked.

        Returns:
            List[Dict[str, str]]: A list of reranked documents in order of relevance.
        """
        if not retrieved_documents:
            logger.warning("No documents provided for reranking.")
            return []

        try:
            model_package_arn = f"arn:aws:bedrock:{self.region}::foundation-model/{self.rerank_model_id}"
            rerank_return_count = len(retrieved_documents)

            # Prepare document sources
            document_sources = [
                {
                    "type": "INLINE",
                    "inlineDocumentSource": {
                        "type": "TEXT",
                        "textDocument": {"text": doc["text"]}
                    }
                }
                for doc in retrieved_documents
            ]

            # Call the Bedrock API
            response = self.bedrock_agent_runtime.rerank(
                queries=[{"type": "TEXT", "textQuery": {"text": input_prompt}}],
                sources=document_sources,
                rerankingConfiguration={
                    "type": "BEDROCK_RERANKING_MODEL",
                    "bedrockRerankingConfiguration": {
                        "numberOfResults": rerank_return_count,
                        "modelConfiguration": {"modelArn": model_package_arn}
                    }
                }
            )

            # Validate response
            results = response.get("results", [])
            if not results:
                logger.error("Reranking failed: No results in response.")
                return []

            # Extract and return reranked documents
            reranked_documents = [
                {"text": retrieved_documents[result["index"]]["text"]}
                for result in results if isinstance(result, dict) and "index" in result
            ]

            return reranked_documents

        except Exception as e:
            logger.error(f"Unexpected error during reranking: {e}")

        return []


from flotorch_core.chunking.hierarical_chunking import HieraricalChunker
from flotorch_core.chunking.fixedsize_chunking import FixedSizeChunker

class ChunkingFactory:
    """
    Factory to create chunking strategies based on configuration.
    """
    @staticmethod
    def create_chunker(chunking_strategy: str, chunk_size: int, chunk_overlap: int, parent_chunk_size: int = None):
        if chunking_strategy.lower() == "hierarchical":
            return HieraricalChunker(chunk_size, chunk_overlap, parent_chunk_size)
        elif chunking_strategy.lower() == "fixed":
            return FixedSizeChunker(chunk_size, chunk_overlap)
        else:
            raise ValueError(f"Unsupported chunking type: {chunking_strategy}")



from typing import Any, Dict, List, Tuple
from flotorch_core.guardrails.guardrails import BaseGuardRail
from flotorch_core.inferencer.inferencer import BaseInferencer


class GuardRailsInferencer(BaseInferencer):
    def __init__(self, base_inferencer: BaseInferencer, base_guardrail: BaseGuardRail):
        self.base_inferencer = base_inferencer
        self.base_guardrail = base_guardrail

    def generate_text(self, user_query: str, context: List[Dict]) -> Tuple[Dict[Any, Any], str]:
        metadata, answer = self.base_inferencer.generate_text(user_query, context)

        guardrail_response = self.base_guardrail.apply_guardrail(answer, 'OUTPUT')
        if guardrail_response['action'] == 'GUARDRAIL_INTERVENED':
            return {
                'guardrail_output_assessment': guardrail_response.get('assessments', []),
                'guardrail_blocked': True
            }, guardrail_response['outputs'][0]['text']
        
        return metadata, answer
    
    def generate_prompt(self, user_query: str, context: List[Dict]) -> str:
        return self.base_inferencer.generate_prompt(user_query, context)

    def format_context(self, context: List[Dict[str, str]]) -> str:
        return self.base_inferencer.format_context(context)


from enum import Enum

class MetricKey(str, Enum):
    CONTEXT_PRECISION = "context_precision"
    ASPECT_CRITIC = "aspect_critic"
    FAITHFULNESS = "faithfulness"
    ANSWER_RELEVANCE = "answer_relevance"



from typing import List, Dict, Tuple, Any
import logging
import time
import random
from .sagemaker_inferencer import SageMakerInferencer

logger = logging.getLogger()
logger.setLevel(logging.INFO)

class LlamaInferencer(SageMakerInferencer):
    def __init__(self, model_id: str, region: str, role_arn: str, n_shot_prompts: int = 0, temperature: float = 0.7, n_shot_prompt_guide_obj: Dict[str, List[Dict[str, str]]] = None):
        super().__init__(model_id, region, role_arn, n_shot_prompts, temperature, n_shot_prompt_guide_obj)
        
    def _prepare_conversation(self, message: str, role: str):
        # Format message and role into a conversation
        if not message or not role:
            logger.error(f"Error in parsing message or role")
        conversation = {
                "role": role, 
                "content": message
            }
        return conversation
    
    def generate_prompt(self, user_query: str, context: List[Dict]) -> Tuple[str, List[Dict[str, Any]]]:
        # Input validation
        if self.n_shot_prompts < 0:
            raise ValueError("n_shot_prompt must be non-negative")
        
        default_prompt = "You are a helpful assistant. Use the provided context to answer questions accurately. If you cannot find the answer in the context, say so"
        # Get system prompt
        system_prompt = default_prompt if not self.n_shot_prompt_guide_obj or not self.n_shot_prompt_guide_obj.get("system_prompt") else self.n_shot_prompt_guide_obj.get("system_prompt")
        
        context_text = ""
        if context:
            context_text = self.format_context(user_query, context)
        
        base_prompt = self.n_shot_prompt_guide_obj.get("user_prompt", "") if self.n_shot_prompt_guide_obj else ""
        
        if self.n_shot_prompts == 0:
            logger.info("into zero shot prompt")
    
            messages = []
            messages.append(self._prepare_conversation(role="user", message=base_prompt))
            if context_text:
                messages.append(self._prepare_conversation(role="user", message=context_text))
            messages.append(self._prepare_conversation(role="user", message=user_query))

            return system_prompt, messages

        # Get examples if nshot is not zero
        examples = self.n_shot_prompt_guide_obj['examples']
        
        # Format examples
        selected_examples = (random.sample(examples, n_shot_prompt) 
                        if len(examples) > n_shot_prompt 
                        else examples)
        
        logger.info(f"into {n_shot_prompt} shot prompt  with examples {len(selected_examples)}")
        
        messages = []
        messages.append(self._prepare_conversation(role="user", message=base_prompt))
        for example in selected_examples:
            if 'example' in example:
                messages.append(self._prepare_conversation(role="user", message=example['example']))
            elif 'question' in example and 'answer' in example:
                messages.append(self._prepare_conversation(role="user", message=example['question']))
                messages.append(self._prepare_conversation(role="assistant", message=example['answer']))
        
        if context_text:
            messages.append(self._prepare_conversation(role="user", message=context_text))
            
        messages.append(self._prepare_conversation(role="user", message=user_query))

        return system_prompt, messages
        
    def construct_payload(self, system_prompt: str, prompt: str) -> dict:
        """
        Constructs llama 4 payload dictionary for model inference with the given prompts and default parameters.
        
        Args:
            system_prompt (str): The system-level prompt that guides the model's behavior
            prompt (str): The actual prompt/query to be sent to the model

        """
        # Define default parameters for the model's generation
        default_params = {
            "max_new_tokens": 256,
            "temperature": self.temperature,
            "top_p": 0.9,
            "do_sample": True
        }
        
        # Prepare payload for model inference
        payload = {
            "system": system_prompt,
            "messages": prompt,
            "parameters": default_params
            }
        
        return payload
    
    def _extract_response(self, response: dict) -> str:
        """
        Parses the response from the model and extracts the generated text.

        Args:
            response (dict): The raw response from the model
        """
        if "choices" in response and isinstance(response["choices"], list):
            return response["choices"][0]["message"]["content"]
        else:
            raise ValueError(f"Unexpected Llama-4 response format: {response}")
        


import os
from opensearchpy import OpenSearch
from flotorch_core.chunking.chunking import Chunk
from flotorch_core.embedding.embedding import BaseEmbedding
from flotorch_core.storage.db.vector.vector_storage import VectorStorage, VectorStorageSearchItem, VectorStorageSearchResponse
from typing import List, Optional

"""
This class is responsible for storing the data in the OpenSearch.
"""

class OpenSearchClient(VectorStorage):
    def __init__(self, host, port, username, password, index, use_ssl=True, verify_certs=False, ssl_assert_hostname=False, ssl_show_warn=False,
                 embedder: Optional[BaseEmbedding] = None):
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.index = index
        self.embedder = embedder
        
        self.client = OpenSearch(
            hosts=[{'host': self.host, 'port': self.port}],
            http_auth=(self.username, self.password),
            use_ssl=use_ssl,
            verify_certs=verify_certs,
            ssl_assert_hostname=ssl_assert_hostname,
            ssl_show_warn=ssl_show_warn,
        )
    
    def write(self, body):
        return self.client.index(index=self.index, body=body)
    
    def read(self, body):
        return self.search(self.index, body)
    
    def write_bulk(self, body: List[dict]):
        return self.client.bulk(body=body)

    # TODO: Need to create a model class for the return type of the search method
    # This model class has to be created in the base class and this return type has to be consitent in all the vector_sotrage classes
    def search(self, chunk: Chunk,  knn: int, hierarchical=False):
        embedding = self.embedder.embed(chunk)
        query_vector = embedding.embeddings
        body = self.embed_query(query_vector, knn, hierarchical)
        response = self.client.search(index=self.index, body=body)

        result = []
        for hit in response['hits']['hits']:
            source = hit['_source']
            result.append(
                VectorStorageSearchItem(
                    execution_id=hit['_id'],
                    chunk_id=source['chunk_id'] if 'chunk_id' in source else None,
                    parent_id=source['parent_id'] if 'parent_id' in source else None,
                    text=source['text'],
                    vectors=source['vectors'],
                    metadata=source['metadata']
                )
            )

        return VectorStorageSearchResponse(
            status=True,
            result=result,
            metadata={
                "embedding_metadata": embedding.metadata
            }
        )
    
    def embed_query(self, query_vector: List[float], knn: int, hierarchical=False):
        vector_field = next((field for field, props in 
                            self.client.indices.get_mapping(index=self.index)[self.index]['mappings']['properties'].items() 
                            if 'type' in props and props['type'] == 'knn_vector'), None)
        query =  {
            "size": knn,
            "query": {
                "knn": {
                    vector_field: {
                        "vector": query_vector,
                        "k": knn
                    }
                }
            },
            "_source": True,
            "fields": ["text", "parent_id"]
        }
        if hierarchical:
            query["collapse"] = {"field": "parent_id.keyword"}

        return query


from abc import ABC, abstractmethod
from typing import List, Dict, Any, Tuple

class BaseInferencer(ABC):
    """
    Abstract base class for all inferencers.
    Defines the common interface and shared functionality for inferencers.
    """

    def __init__(self, model_id: str, region: str = "us-east-1", n_shot_prompts: int = 0, temperature: float = 0.7, n_shot_prompt_guide_obj: Dict[str, List[Dict[str, str]]] = None):
        """
        Initialize the inferencer with required parameters.

        Args:
            model_id (str): Identifier for the model.
            region (str): AWS region or equivalent for the inferencing service.
            n_shot_prompts (int): Number of examples to include in few-shot learning. Defaults to 0.
            temperature (float): Sampling temperature for response generation. Defaults to 0.7.
            n_shot_prompt_guide_obj (Any): Guide object for few-shot examples. Defaults to None.
        """
        self.model_id = model_id
        self.region_name = region
        self.n_shot_prompts = n_shot_prompts
        self.temperature = temperature
        self.n_shot_prompt_guide_obj = n_shot_prompt_guide_obj

    @abstractmethod
    def generate_text(self, user_query: str, context: List[Dict]) -> Tuple[Dict[Any, Any], str]:
        """
        Generate a response based on the user query and context.

        Args:
            user_query (str): The question or input from the user.
            context (List[Dict]): Contextual data to assist in generating a response.
            default_prompt (str): Default prompt to guide the response generation.
            **kwargs: Additional parameters for customization.

        Returns:
            Tuple[Dict[Any, Any], str]: Metadata and the generated response text.
        """
        pass

    @abstractmethod
    def generate_prompt(self, user_query: str, context: List[Dict]) -> str:
        """
        Construct a prompt for the inferencer based on the user query and context.

        Args:
            default_prompt (str): Default prompt to guide the response generation.
            user_query (str): The question or input from the user.
            context (List[Dict]): Contextual data to assist in generating a response.

        Returns:
            str: The constructed prompt.
        """
        pass

    @abstractmethod
    def format_context(self, context: List[Dict[str, str]]) -> str:
        """
        Format context documents into a single string for inclusion in the prompt.

        Args:
            context (List[Dict[str, str]]): List of context documents.

        Returns:
            str: Formatted context as a single string.
        """
        pass

    def _extract_response(self, response: Dict) -> str:
        """
        Extract the generated text from the response.

        Args:
            response (Dict): Response from the inferencing service.

        Returns:
            str: Extracted text from the response.
        """
        raise NotImplementedError("Subclasses must implement _extract_response")



from .config_provider import ConfigProvider

class Config:
    """
    Main configuration class that uses a configuration provider.
    """

    def __init__(self, provider: ConfigProvider):
        """
        Initializes the Config class with the specified provider.
        Args:
            provider (ConfigProvider): The configuration provider.
        """
        self.provider = provider

    def get_region(self) -> str:
        """
        Retrieves the AWS region from the configuration provider.
        Returns:
            str: The AWS region.
        Raises:
            ValueError: If the region is not set.
        """
        region = self.provider.get("AWS_REGION", "us-east-1")
        if not region:
            raise ValueError("AWS region is not set. Value not presnt in configuration")
        return region
    
    def get_opensearch_host(self) -> str:
        """
        Retrieves the OpenSearch host from the configuration provider.
        """
        open_search_host = self.provider.get("OPENSEARCH_HOST", "localhost")
        if not open_search_host:
            raise ValueError("OpenSearch host is not set. Value not presnt in configuration")
        return open_search_host

    def get_opensearch_port(self) -> int:
        """
        Retrieves the OpenSearch port from the configuration provider.
        """
        open_search_port = int(self.provider.get("OPENSEARCH_PORT", 443))
        if not open_search_port:
            raise ValueError("OpenSearch port is not set. Value not presnt in configuration")
        return open_search_port
    
    def get_opensearch_username(self) -> str:
        """
        Retrieves the OpenSearch username from the configuration provider.
        """
        open_search_username = self.provider.get("OPENSEARCH_USERNAME")
        if not open_search_username:
            raise ValueError("OpenSearch username is not set. Value not presnt in configuration")
        return open_search_username

    def get_opensearch_password(self) -> str:
        """
        Retrieves the OpenSearch password from the configuration provider.
        """
        open_search_password = self.provider.get("OPENSEARCH_PASSWORD")
        if not open_search_password:
            raise ValueError("OpenSearch password is not set. Value not presnt in configuration")
        return open_search_password

    def get_opensearch_index(self) -> str:
        """
        Retrieves the OpenSearch index name from the configuration provider.
        """
        open_search_index = self.provider.get("OPENSEARCH_INDEX")
        if not open_search_index:
            raise ValueError("OpenSearch index is not set. Value not presnt in configuration")
        return open_search_index

    def get_task_token(self) -> str:
        """
        Retrieves the task token from the configuration provider.
        """
        task_token = self.provider.get("TASK_TOKEN")
        if not task_token:
            raise ValueError("task token is not set. Value not presnt in configuration")
        return task_token

    def get_fargate_input_data(self) -> str:
        """
        Retrieves the input data for fargate handlers from the configuration provider.
        """
        input_data = self.provider.get("INPUT_DATA", {})
        if not input_data:
            raise ValueError("input data is not set. Value not present in configuration")
        return input_data
    
    def get_experiment_table_name(self) -> str:
        """
        Retrieves the experiment table name from the configuration provider.
        """
        experiment_table_name = self.provider.get("experiment_table")
        if not experiment_table_name:
            raise ValueError("experiment table name is not set. Value not present in configuration")
        return experiment_table_name
    
    def get_execution_table_name(self) -> str:
        """
        Retrieves the execution table name from the configuration provider.
        """
        execution_table_name = self.provider.get("execution_table")
        if not execution_table_name:
            raise ValueError("execution table name is not set. Value not present in configuration")
        return execution_table_name
    
    def get_experiment_question_metrics_table(self) -> str:
        """
        Retrieves the experiment question metrics table name from the configuration provider.
        """
        experiment_question_metrics_table = self.provider.get("experiment_question_metrics_table")
        if not experiment_question_metrics_table:
            raise ValueError("experiment question metrics table name is not set. Value not present in configuration")
        return experiment_question_metrics_table
    
    def get_execution_model_invocations_table(self) -> str:
        """
        Retrieves the execution model invocations table name from the configuration provider.
        """
        execution_model_invocations_table = self.provider.get("execution_model_invocations_table")
        if not execution_model_invocations_table:
            raise ValueError("execution model invocations table name is not set. Value not present in configuration")
        return execution_model_invocations_table
    
    def get_sagemaker_arn_role(self) -> str:
        """
        Retrieves the SageMaker ARN role from the configuration provider.
        """
        sagemaker_arn_role = self.provider.get("sagemaker_role_arn")
        if not sagemaker_arn_role:
            raise ValueError("sagemaker arn role is not set. Value not present in configuration")
        return sagemaker_arn_role
    
    def get_experimentid_index(self) -> str:
        """
        Retrieves the experiment ID index name from the configuration provider.
        """
        experiment_id_index = self.provider.get("experiment_question_metrics_experimentid_index")
        if not experiment_id_index:
            raise ValueError("experiment id index is not set. Value not present in configuration")
        return experiment_id_index
    
    def get_postgres_db(self) -> str:
        """
        Retrieves the Postgres database name from the configuration provider.
        """
        postgres_db = self.provider.get("postgres_db_name")
        if not postgres_db:
            raise ValueError("Postgres database is not set. Value not present in configuration")
        return postgres_db
    
    def get_postgres_user(self) -> str:
        """
        Retrieves the Postgres user from the configuration provider.
        """
        postgres_user = self.provider.get("postgres_user")
        if not postgres_user:
            raise ValueError("Postgres user is not set. Value not present in configuration")
        return postgres_user
    
    def get_postgres_password(self) -> str:
        """
        Retrieves the Postgres password from the configuration provider.
        """
        postgres_password = self.provider.get("postgres_password")
        if not postgres_password:
            raise ValueError("Postgres password is not set. Value not present in configuration")
        return postgres_password
    
    def get_postgres_host(self) -> str:
        """
        Retrieves the Postgres host from the configuration provider.
        """
        postgres_host = self.provider.get("postgres_host")
        if not postgres_host:
            raise ValueError("Postgres host is not set. Value not present in configuration")
        return postgres_host

    def get_postgres_port(self) -> int:
        """
        Retrieves the Postgres port from the configuration provider.
        """
        postgres_port = int(self.provider.get("postgres_port", 5432))
        if not postgres_port:
            raise ValueError("Postgres port is not set. Value not present in configuration")
        return postgres_port
    
    def get_db_type(self) -> str:
        """
        Retrieves the database type from the configuration provider.
        """
        db_type = self.provider.get("db_type")
        if not db_type:
            raise ValueError("db type is not set. Value not present in configuration")
        return db_type




from typing import Dict, List
from flotorch_core.inferencer.bedrock_inferencer import BedrockInferencer
from flotorch_core.inferencer.gateway_inferencer import GatewayInferencer
from flotorch_core.inferencer.inferencer import BaseInferencer
from flotorch_core.inferencer.sagemaker_inferencer import SageMakerInferencer
from flotorch_core.inferencer.llama_inferencer import LlamaInferencer


class InferencerProviderFactory:
    """
    Factory to create inferencer based on the service name.
    """
    @staticmethod
    def create_inferencer_provider(gateway_enabled: bool, base_url: str, api_key: str, service: str, model_id: str, region: str, arn_role: str, n_shot_prompts: int = 0, temperature: float = 0.7, n_shot_prompt_guide_obj: Dict[str, List[Dict[str, str]]] = None) -> BaseInferencer:
        if gateway_enabled:
            return GatewayInferencer(
                model_id=model_id, 
                api_key=api_key, 
                base_url=base_url, 
                n_shot_prompts=n_shot_prompts, 
                n_shot_prompt_guide_obj=n_shot_prompt_guide_obj
            )
        
        if service == 'bedrock':
            return BedrockInferencer(model_id, region, n_shot_prompts, temperature, n_shot_prompt_guide_obj)
        elif service == 'sagemaker':
            if model_id.startswith("meta-vlm-llama-4"):
                return LlamaInferencer(model_id, region, arn_role, n_shot_prompts, temperature, n_shot_prompt_guide_obj)
            else:
                return SageMakerInferencer(model_id, region, arn_role, n_shot_prompts, temperature, n_shot_prompt_guide_obj)
        else:
            raise ValueError(f"Unsupported service scheme: {service}")



import logging
import traceback
import requests
from .storage import StorageProvider
from typing import Generator, Any, List, Union, Optional, Dict
from pydantic import BaseModel, Field, ValidationError, parse_obj_as


logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

class UrlStorageProvider(StorageProvider):
    """
    Url storage provider
    """
    def __init__(self):
        super().__init__()

    def get_path(self, uri: str) -> str:
        """
        For HTTP, the entire URI is essentially the "path" to the resource.
        """
        return uri

    def write(self, path: str, data: Union[Dict[str, Any], List[Any], Any], headers: Optional[Dict[str, str]] = None) -> None:
        """
        Sends data as a JSON payload to the specified API path with optional custom headers.

        Args:
            path (str): The API endpoint URL.
            data (Union[Dict[str, Any], List[Any], Any]): The data to send.
                                                            This can be a dictionary, list,
                                                            or any other JSON-serializable Python object.
            headers (Optional[Dict[str, str]]): Optional dictionary of HTTP headers to send with the request.
                                                 If not provided, 'Content-Type: application/json' will be
                                                 automatically set by 'requests' due to the 'json' parameter.
                                                 If provided, these headers will be used, and 'Content-Type'
                                                 will be added if not already present.
        """
        try:
            logger.info(f"Sending data to API URL: {path} with payload: {data}")
            response = requests.post(path, json=data, headers=headers)

            if response.status_code == 200:
                logger.info(f"Successfully sent data to API URL: {path} with status code: {response.status_code}")
            elif response.status_code == 400:
                error_message = f"API Error (400 Bad Request) for {path}: Invalid payload. Response: {response.text}"
                logger.error(error_message)
                raise requests.exceptions.RequestException(error_message)
            elif response.status_code == 404:
                error_message = f"API Error (404 Not Found) for {path}: Invalid project/experiment UUIDs. Response: {response.text}"
                logger.error(error_message)
                raise requests.exceptions.RequestException(error_message)
            else:
                response.raise_for_status()

        except requests.exceptions.RequestException as e:
            traceback.print_exc()
            logger.error(f"Error sending data to API URL {path}: {e}")
            raise
        except Exception as e:
            traceback.print_exc()
            logger.error(f"An unexpected error occurred: {e}")
            raise

    def read(self, path: str, headers: Optional[Dict[str, str]] = None) -> Generator[bytes, None, None]:
        """
        Reads data from the specified HTTP(S) URL.
        Args:
            path (str): The URL to read the data from.
            headers (Optional[Dict[str, str]]): Optional dictionary of HTTP headers to send with the request.
        Returns:
            Generator[bytes, None, None]: A generator that yields the data read.
        """
        logger.info(f"Attempting to read from URL: {path}")
        try:
            response = requests.get(path, headers=headers)
            response.raise_for_status()
            yield response.content
        except requests.exceptions.RequestException as e:
            logger.error(f"Error reading from HTTP(S) path {path}: {e}")
            raise